# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for a base class for configurations """

import ast
import configparser
import os
from abc import ABCMeta, abstractmethod


SECTION = "section"
OPTION  = "option"
TYPE    = "type"
DEFAULT = "default"
HELP    = "help"

ERROR_ATTRIBUTE = "There is no attribute '{}'"
ERROR_FILE      = "Configuration file '{}' does not exists"
ERROR_MISSING   = "Missing {} '{}' in configuration file '{}'"
ERROR_FORMAT    = "Configuration value is not well formatted"
ERROR_TYPE      = "The value is of type '{}' and not the intended type '{}'"

FEASIBLE_SCALAR_TYPES = (int, float, str, bool)
FEASIBLE_TYPES = (int, float, str, bool, dict, list)

ERROR_SCALAR = "Values must be of one of the types int, float, str or bool, not '{}'"
ERROR_NESTED = "Values must be of one of the types int, float, str, bool, list or dict, not '{}'"
ERROR_KEYS   = "Keys must be of type str not '{}'"
ERROR_LEVEL  = "Level of nested dictionaries exceeds specified maximal depth {}"


class Config(dict):
    """
    A configuration is a container to store several values.
    This serves as a base class for the implementation of different configurations.
    """
    __metaclass__ = ABCMeta

    def __init__(self, set_defaults=False, **kwargs):
        """
         initialize configuration object

        :param set_defaults: if True, sets default values to the configuration options, otherwise None is set
        """
        super().__init__(**kwargs)
        if set_defaults:
            self.set_defaults(False)

    def __getattr__(self, option):
        """
        get the value of the option of the configuration,
        where attributes and items of the internal dictionary are handled equivalently,
        thus we have `cfg.__getattr__('option') == cfg.option == cfg['option']`
        """
        return self[option]

    def __setattr__(self, option, value):
        """ change the value of the option of the configuration """
        self[option] = value

    def __delattr__(self, option):
        """
        delete the option from the configuration, resets value to None,
        raises an exception if the option does not exist
        """
        del self[option]

    def __getitem__(self, option):
        """ get the value of a specific option of the configuration """
        if option in self:
            return super().__getitem__(option)
        raise AttributeError(ERROR_ATTRIBUTE.format(option))

    def __delitem__(self, option):
        """ reset the option to None, raises an exception if the option does not exist """
        if option in self:
            self[option] = None
        else:
            raise AttributeError(ERROR_ATTRIBUTE.format(option))

    @classmethod
    def get_name(cls):
        """ get the name of the configuration """
        return cls.__name__

    @staticmethod
    @abstractmethod
    def get_options_dict():
        """
        get the dictionary of options and additional information,
        needs to be implemented in inheriting class
        """
        raise NotImplementedError()

    def update(self, **kwargs):
        """ update the given options, raises an exception if some option does not exist """
        for option in kwargs:
            if option not in self.get_options_dict().keys():
                raise AttributeError(ERROR_ATTRIBUTE.format(option))
        super().update(kwargs)

    def set_defaults(self, override=True):
        """
        set the values to the default ones,
        defaults need to be specified by an implementation of the `get_options_dict` method
        """
        kwargs = self.get_defaults()
        if not override:
            for key in list(kwargs):
                if self[key] is not None:
                    kwargs.pop(key)
        else:
            for key in list(kwargs):
                if kwargs[key] is None:
                    kwargs.pop(key)
        self.update(**kwargs)

    def to_string(self, max_depth=4, omit_keys=None):
        """
        get a unique string from the configuration,
        can only be applied, if the configuration contains the suitable entries

        :param (int) max_depth: the maximum number of the recursive resolution of dicts in dicts,
                                gives a warning and incomplete result if the configuration exceeds it
        :param (list) omit_keys: list of strings providing the keys that should be omitted during the string conversion,
                                 the corresponding key-value pairs will be omitted on all depth levels in case we have
                                 nested dictionaries
        :returns: unique string from configuration file if possible, None otherwise
        """
        return dict_to_string(self, max_depth=max_depth, omit_keys=omit_keys)

    def save(self, filename):
        """
        save the configuration in a file

        :param (str) filename: path to the file where to store the configuration file in
        """
        with open(filename, "w", encoding="utf-8") as config_file:
            cfg_parser = configparser.ConfigParser()
            section = self.get_name()
            options = self.get_options_dict().keys()
            cfg_parser.add_section(section)
            for option in options:
                cfg_parser.set(section=section, option=str(option), value=str(self[option]))
            cfg_parser.write(config_file)

    @classmethod
    def load(cls, filename):
        """
        load configuration from a file

        :param (str) filename: path to configuration file
        :returns: the loaded configuration
        """
        if not os.path.exists(filename):
            raise ValueError(ERROR_FILE.format(filename))
        cfg_name = cls.get_name()
        options = cls.get_options_dict()
        config_parser = configparser.ConfigParser()
        config_parser.read(filename)

        if not config_parser.has_section(cfg_name):
            raise ValueError(ERROR_MISSING.format(SECTION, cfg_name, filename))
        init_kwargs = {}
        for option in options:
            if not config_parser.has_option(cfg_name, option):
                raise ValueError(ERROR_MISSING.format(OPTION, option, filename))
            intended_type = options[option].get(TYPE)
            init_kwargs[option] = cls._cast_type(config_value=config_parser.get(cfg_name, option),
                                                 intended_type=intended_type)
        return cls(**init_kwargs)

    @classmethod
    def _cast_type(cls, config_value, intended_type):
        """
        cast the loaded elements of the configuration to the correct type

        :return: cast value of given config_value or None if there is a missmatch between intended type and actual type
        """
        try:
            config_value = ast.literal_eval(config_value)
        except (ValueError, SyntaxError):
            assert isinstance(config_value, str), ERROR_FORMAT
        if isinstance(config_value, intended_type):
            return config_value
        if config_value is None:
            return None
        if intended_type is str:
            return str(config_value)
        raise TypeError(ERROR_TYPE.format(type(config_value), intended_type))

    @classmethod
    def get_defaults(cls):
        """ get default values specified in `get_options_dict` """
        return {name: d.get(DEFAULT, None) for name, d in cls.get_options_dict().items()}

    @classmethod
    def get_help(cls):
        """ get description/help that is specified in 'get_options_dict' """
        cfg_name = cls.get_name()
        options = cls.get_options_dict()
        n = max(len(name) for name in options.keys()) + 1
        lines = ""
        for name in options:
            help_str = options[name].get(HELP, "")
            lines += "\t" + name + " " * (n - len(name)) + f"= <{help_str}>\n"
        return f"configuration file with the following entries:\n\t[{cfg_name}]\n{lines}"


def dict_to_string(dictionary, max_depth, omit_keys=None, _depth_counter=0):
    """
    get a unique string from an ordered dictionary if possible,
    the keys of the dictionary can be dictionaries themselves,
    in this case they are recursively resolved up until a maximal depth

    :param (dict) dictionary: dictionary to be formatted
    :param (int) max_depth: maximum level of nested dicts that should be resolved, if exceeded, return None
    :param (list or None) omit_keys: list of keys that should be omitted during the string conversion,
                                     the corresponding key-value pairs will be omitted on all depth levels in case we
                                     have nested dictionaries
    :param (int) _depth_counter: counting the levels of nested dictionaries that have been resolved
    :return: unique string if formatting is possible
    """
    dict_string = ""
    for key, value in dictionary.items():
        if omit_keys and key in omit_keys:
            continue
        if not isinstance(key, str):
            raise ValueError(ERROR_KEYS.format(type(key).__name__))
        if not isinstance(value, FEASIBLE_TYPES):
            raise ValueError(ERROR_NESTED.format(type(value).__name__))
        if isinstance(value, dict):
            if _depth_counter >= max_depth:
                raise ValueError(ERROR_LEVEL.format(max_depth))
            substring = dict_to_string(value, max_depth, omit_keys=omit_keys, _depth_counter=_depth_counter + 1)
            dict_string += f"_{_to_string(key)}{substring}"
        elif isinstance(value, list):
            substring = "".join(f"_{_to_string(element)}" for element in value)
            dict_string += f"_{_to_string(key)}{substring}"
        elif isinstance(value, str):
            dict_string += f"_{_to_string(key)}_{_to_string(value)}"
        else:
            dict_string += f"_{_to_string(key)}{_to_string(value)}"
    return dict_string

def _to_string(value):
    """
     get a unique string to a valid input value

    :param value: needs to be int, float, string or bool
    """
    if isinstance(value, str):
        return value
    if isinstance(value, bool):
        return str(value)
    if isinstance(value, (int, float)):
        return f'{value:e}'
    raise ValueError(ERROR_SCALAR.format(type(value).__name__))



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for Solution """

from numbers import Integral
from functools import partial

from .variable_mapping import VariableMapping
from .utils.variables import are_consecutive, to_string


ERROR_NEITHER     = "Solution is neither binary nor ising"
ERROR_MAPPING     = "Variable '{}' does not have a mapping to a new variable"
ERROR_EMPTY       = "Solution is empty, there are no variables to be replaced"
ERROR_VARIABLES   = "Number of provided variables is different than of contained variables"
ERROR_COMPACT     = "Solution is not compact"
ERROR_REPLACEMENT = "Dictionary of variables or VariableMapping is expected"

ADDITIONAL_INFO_ATTRS = ["objective_value",
                         "solving_success",
                         "solving_status",
                         "solving_time",
                         "total_time",
                         "dual_gap",
                         "dual_bound"]


class Solution(dict):
    """ A solution is a mapping from variables to values in the domain of the corresponding problem. """

    def __init__(self,
                 var_assignments=None,
                 objective_value=None,
                 solving_success=False,
                 solving_status=None,
                 solving_time=0,
                 total_time=None,
                 dual_gap=None,
                 dual_bound=None,
                 name=None):
        """
        initialize Solution object

        :param (dict or list) var_assignments: dictionary or list containing the solution,
                                               i.e., a mapping from variables to values in the corresponding domain,
                                               e.g., {('x', 1, 2): 1, ('x', 1, 1): 0},
                                               a list of values is considered as mapping from flat variables
        :param (float or None) objective_value: objective value or current best value returned by the solver
        :param (bool) solving_success: True if solver successfully optimized the model
        :param (str or None) solving_status: status of the solver at the end of the optimization, e.g. 'timeout'
        :param (float or None) solving_time: runtime of the optimization
        :param (float or None) total_time: time since model object was created
        :param (float or None) dual_gap: gap to the dual solution
        :param (float or None) dual_bound: value of the dual solution
        :param (str or None) name: identifying name to differ between several solutions
        """
        if isinstance(var_assignments, list):
            var_assignments = dict(enumerate(var_assignments))
        super().__init__(var_assignments or {})

        self.objective_value = objective_value
        self.solving_success = solving_success
        self.solving_status = solving_status
        self.solving_time = solving_time
        self.total_time = total_time
        self.dual_gap = dual_gap
        self.dual_bound = dual_bound
        self.name = name

    def __eq__(self, other):
        # we also allow for the comparison with plain dicts where the name and the additional info is irrelevant
        if isinstance(other, dict) and super().__eq__(other):
            if isinstance(other, Solution):
                return self.name == other.name and self._get_additional_info_dict() == other._get_additional_info_dict()
            return True
        return False

    def __ne__(self, other):
        # needs to be implemented otherwise would just use dict comparison
        return not self.__eq__(other)

    def __repr__(self):
        """ get nice string representation """
        self.sort_entries()
        return self.__class__.__name__ + f"({super().__repr__()}, {self.objective_value}, {repr(self.name)})"

    def __str__(self):
        """ convert the solution to a human-readable string """
        self.sort_entries()
        return "\n".join(to_string(var) + f" \t= {value}" for var, value in self.items())

    def sort_entries(self):
        """ does not change the solution but the ordering of its entries in the dict, mainly for nice output """
        sorted_items = {key: self[key] for key in sorted(self.keys())}
        super().clear()
        super().update(sorted_items)

    def has_only_binary_values(self):
        """ check if the solution has only binary values"""
        return all(v in (0, 1) for v in self.values())

    def has_only_ising_values(self):
        """ check if the solution hase only ising values"""
        return all(v in (-1, 1) for v in self.values())

    def apply_to_var_assignments(self, func, new_name=None):
        """
        apply the function to the variable assignments and thus extract another solution

        :param (callable) func: a mapping from the dictionary of variable assignments to a new one
        :param (str or None) new_name: name of the new solution, by default the original name is used
        :return: the new solution
        """
        return Solution(func(self), name=new_name or self.name, **self._get_additional_info_dict())

    def _get_additional_info_dict(self):
        return {attr: getattr(self, attr) for attr in ADDITIONAL_INFO_ATTRS}

    def to_ising(self, invert=False, new_name=None):
        """
        transform the solution into an Ising solution by applying the transformation (2x-1) resp. (-2x+1)
        or, if the solution is already an Ising solution, with invert=True it can be inverted

        :param (bool) invert: if True, the resulting Ising will additionally be inverted (1->-1, -1->1)
        :param (str or None) new_name: name of the new solution, by default the original name is used
        :return: the new solution with Ising values
        """
        if self.has_only_ising_values():
            if not invert:
                return self
            return self.apply_to_var_assignments(invert_ising, new_name)
        if self.has_only_binary_values():
            return self.apply_to_var_assignments(partial(binary_to_ising, invert=invert), new_name)
        raise ValueError(ERROR_NEITHER)

    def to_binary(self, is_inverted=False, new_name=None):
        """
        transform the solution into a binary solution by applying the transformation (x+1)/2 resp. (-x+1)/2

        :param (bool) is_inverted: if True, the given Ising solution is from an inverted Ising objective,
                                   therefore it gets inverted before the conversion to Ising
        :param (str or None) new_name: name of the new solution, by default the original name is used
        :return: the new solution with binary values
        """
        if self.has_only_binary_values():
            return self
        if self.has_only_ising_values():
            return self.apply_to_var_assignments(partial(ising_to_binary, is_inverted=is_inverted), new_name)
        raise ValueError(ERROR_NEITHER)

    def decompact(self, variables, new_name=None):
        """
        replace the variables in the solution with the original (non-compact) variables

        :param (list or dict or VariableMapping) variables: the original variables from the non-compact polynomial
        :param (str or None) new_name: name of the new solution, by default the original name is used
        :return: the solution with non-compact variables
        """
        if not self:
            raise ValueError(ERROR_EMPTY)
        if not len(self) == len(variables):
            raise ValueError(ERROR_VARIABLES)
        if not are_consecutive(set(self.keys())):
            raise ValueError(ERROR_COMPACT)
        return self.replace_variables(variables, new_name=new_name, check_all=False)

    def replace_variables(self, replacement, new_name=None, check_all=True):
        """
        replace the variables in the solution with different ones

        :param (list or dict or VariableMapping) replacement: mapping to the new variables,
                                                              if replacement is a dictionary, it should be of format
                                                                   {old_var : new_var},
                                                              can be a list if the polynomial is flat
        :param (str or None) new_name: name of the new solution, by default the original name is used
        :param (bool) check_all: if False, those variables which do not have a mapping will not be replaced
                                     and no error is thrown
                                 if True, will throw error if there is a variable which does not have a mapping
        :return: the solution with replaced variables
        """
        if isinstance(replacement, list):
            replacement = dict(enumerate(replacement))
        if not isinstance(replacement, (dict, VariableMapping)):
            raise ValueError(ERROR_REPLACEMENT)
        return self.apply_to_var_assignments(partial(_replace, replacement=replacement, check_all=check_all), new_name)


def binary_to_ising(var_assignment, invert=False):
    """
    convert the variable assignment to the one of the corresponding Ising problem

    :param (dict) var_assignment: solution to a binary problem as a mapping from variables to values
    :param (bool) invert: if True, convert binary to inverted Ising
    :return: the solution to the corresponding Ising problem
    """
    return {k: (-2*v+1 if invert else 2*v-1) for k, v in var_assignment.items()}

def ising_to_binary(var_assignment, is_inverted=False):
    """
    convert the variable assignment to the one of the corresponding binary problem

    :param (dict) var_assignment: solution to an Ising problem as a mapping from variables to values
    :param (bool) is_inverted: if True, the var_assignment is a solution of an inverted Ising problem and will thus be
                               inverted beforehand
    :return: the solution to the corresponding binary problem
    """
    return {k: ((int((-v+1)/2) if isinstance(v, Integral) else (-v+1)/2) if is_inverted
                else (int((v+1)/2)) if isinstance(v, Integral) else (v+1)/2)  # if was integer, keep integer
            for k, v in var_assignment.items()}

def invert_ising(var_assignment):
    """
    convert the variable assignment to the one of an inverted Ising problem or vice versa

    :param (dict) var_assignment: solution to an Ising or inverted Ising problem as a mapping from variables to values
    :return: the solution to the corresponding inverted Ising problem
    """
    return {k: -v for k, v in var_assignment.items()}

def _replace(var_assignment, replacement, check_all):
    # here replacement is a dictionary of the format {old_var : new_var}
    try:
        return {replacement[old_var] if check_all else replacement.get(old_var, old_var): coefficient
                for old_var, coefficient in var_assignment.items()}
    except KeyError as ke:
        raise ValueError(ERROR_MAPPING.format(ke.args[0])) from ke



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ExampleObjectiveTerms """

from quark import ObjectiveTerms, PolyBinary


ONE_COLOR_PER_NODE = "one_color_per_node"
COLORED_EDGES      = "colored_edges"
X = "x"


class ExampleObjectiveTerms(ObjectiveTerms):
    """
    An example implementation of an ObjectiveTerms for testing based on the MaxColorableSubgraph problem

    The binary variable x_n_c is 1 if node n is colored with color c and 0 otherwise.
    """

    @staticmethod
    def _get_constraint_terms_names():
        """
        get the names of the objective terms whose value must be zero to have the corresponding constraints fulfilled,
        which need to be a subset of all objective terms names
        """
        return [ONE_COLOR_PER_NODE]

    @staticmethod
    def _get_objective_terms(instance):
        """ get the objective terms from the instance data """
        objective_terms = {COLORED_EDGES: PolyBinary(),
                           ONE_COLOR_PER_NODE : PolyBinary()}

        # actual objective
        # counting the number of wrongly colored edges:
        # sum_[c in Colors] sum_[(n, m) in Edges] (1 * x_n_c * x_m_c)
        for color in instance.colors:
            objective_terms[COLORED_EDGES] += PolyBinary({((X, node1, color), (X, node2, color)): 1
                                                          for node1, node2 in instance.edges})

        # objective terms from the constraints
        # every node should get exactly one color:
        # sum_[n in Nodes] (sum_[c in Colors] x_n_c - 1)^2
        for node in instance.nodes:
            poly = PolyBinary({((X, node, color),): 1 for color in instance.colors}) - 1
            objective_terms[ONE_COLOR_PER_NODE] += poly * poly

        return objective_terms



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the ObjectiveTerms by inherited example and without """

import pytest

from quark import Polynomial, PolyBinary, PolyIsing, ObjectiveTerms
from quark.testing import ExampleInstance, ExampleObjectiveTerms
from quark.testing.example_objective_terms import X, COLORED_EDGES, ONE_COLOR_PER_NODE
from quark.objective_terms import get_naive_terms_weights


B, G, R = "blue", "green", "red"
REDUCTION = "reduction"

@pytest.fixture(name="objective_terms_different_types")
def fixture_objective_terms_different_types():
    """ provide test objective terms with different types """
    poly_ising = PolyIsing({(0, 1): 2, (2, 3): 4})
    poly_binary = PolyBinary({(0, 1): 2, (2, 3): 4})
    poly_none = Polynomial({(1, 2): 3})
    terms = {"ising": poly_ising, "binary": poly_binary, "none": poly_none}
    yield ExampleObjectiveTerms(objective_terms=terms, constraint_terms_names=["ising", "binary"])


def test_init(objective_terms, exp_objective_terms):
    """ test objective terms creation """
    assert objective_terms == exp_objective_terms

def test_repr(objective_terms):
    """ test string representation """
    exp_str = "ObjectiveTerms({'colored_edges': PolyBinary({(('x', 1, 'blue'), ('x', 2, 'blue')): 1, " \
                                                           "(('x', 1, 'blue'), ('x', 3, 'blue')): 1, " \
                                                           "(('x', 1, 'blue'), ('x', 5, 'blue')): 1, " \
                                                           "(('x', 1, 'green'), ('x', 2, 'green')): 1, " \
                                                           "(('x', 1, 'green'), ('x', 3, 'green')): 1, " \
                                                           "(('x', 1, 'green'), ('x', 5, 'green')): 1, " \
                                                           "(('x', 1, 'red'), ('x', 2, 'red')): 1, " \
                                                           "(('x', 1, 'red'), ('x', 3, 'red')): 1, " \
                                                           "(('x', 1, 'red'), ('x', 5, 'red')): 1, " \
                                                           "(('x', 2, 'blue'), ('x', 3, 'blue')): 1, " \
                                                           "(('x', 2, 'green'), ('x', 3, 'green')): 1, " \
                                                           "(('x', 2, 'red'), ('x', 3, 'red')): 1, " \
                                                           "(('x', 3, 'blue'), ('x', 4, 'blue')): 1, " \
                                                           "(('x', 3, 'blue'), ('x', 5, 'blue')): 1, " \
                                                           "(('x', 3, 'green'), ('x', 4, 'green')): 1, " \
                                                           "(('x', 3, 'green'), ('x', 5, 'green')): 1, " \
                                                           "(('x', 3, 'red'), ('x', 4, 'red')): 1, " \
                                                           "(('x', 3, 'red'), ('x', 5, 'red')): 1, " \
                                                           "(('x', 4, 'blue'), ('x', 5, 'blue')): 1, " \
                                                           "(('x', 4, 'green'), ('x', 5, 'green')): 1, " \
                                                           "(('x', 4, 'red'), ('x', 5, 'red')): 1}), " \
              "'one_color_per_node': PolyBinary({(): 5, " \
                              "(('x', 1, 'blue'),): -1, (('x', 1, 'green'),): -1, (('x', 1, 'red'),): -1, " \
                              "(('x', 2, 'blue'),): -1, (('x', 2, 'green'),): -1, (('x', 2, 'red'),): -1, " \
                              "(('x', 3, 'blue'),): -1, (('x', 3, 'green'),): -1, (('x', 3, 'red'),): -1, " \
                              "(('x', 4, 'blue'),): -1, (('x', 4, 'green'),): -1, (('x', 4, 'red'),): -1, " \
                              "(('x', 5, 'blue'),): -1, (('x', 5, 'green'),): -1, (('x', 5, 'red'),): -1, " \
                              "(('x', 1, 'blue'), ('x', 1, 'green')): 2, (('x', 1, 'blue'), ('x', 1, 'red')): 2, " \
                              "(('x', 1, 'green'), ('x', 1, 'red')): 2, (('x', 2, 'blue'), ('x', 2, 'green')): 2, " \
                              "(('x', 2, 'blue'), ('x', 2, 'red')): 2, (('x', 2, 'green'), ('x', 2, 'red')): 2, " \
                              "(('x', 3, 'blue'), ('x', 3, 'green')): 2, (('x', 3, 'blue'), ('x', 3, 'red')): 2, " \
                              "(('x', 3, 'green'), ('x', 3, 'red')): 2, (('x', 4, 'blue'), ('x', 4, 'green')): 2, " \
                              "(('x', 4, 'blue'), ('x', 4, 'red')): 2, (('x', 4, 'green'), ('x', 4, 'red')): 2, " \
                              "(('x', 5, 'blue'), ('x', 5, 'green')): 2, (('x', 5, 'blue'), ('x', 5, 'red')): 2, " \
                              "(('x', 5, 'green'), ('x', 5, 'red')): 2})}, ['one_color_per_node'], None)"
    assert repr(objective_terms) == exp_str

def test_str(objective_terms):
    """ test string representation """
    exp_str = "min  P_0 * ( +1 x_1_blue x_2_blue +1 x_1_blue x_3_blue +1 x_1_blue x_5_blue +1 x_1_green x_2_green " \
                           "+1 x_1_green x_3_green +1 x_1_green x_5_green +1 x_1_red x_2_red +1 x_1_red x_3_red " \
                           "+1 x_1_red x_5_red +1 x_2_blue x_3_blue +1 x_2_green x_3_green +1 x_2_red x_3_red " \
                           "+1 x_3_blue x_4_blue +1 x_3_blue x_5_blue +1 x_3_green x_4_green +1 x_3_green x_5_green " \
                           "+1 x_3_red x_4_red +1 x_3_red x_5_red +1 x_4_blue x_5_blue +1 x_4_green x_5_green " \
                           "+1 x_4_red x_5_red )\n" \
              "     + P_1 * ( +5 -1 x_1_blue -1 x_1_green -1 x_1_red -1 x_2_blue -1 x_2_green -1 x_2_red -1 x_3_blue " \
                             "-1 x_3_green -1 x_3_red -1 x_4_blue -1 x_4_green -1 x_4_red -1 x_5_blue -1 x_5_green " \
                             "-1 x_5_red +2 x_1_blue x_1_green +2 x_1_blue x_1_red +2 x_1_green x_1_red " \
                             "+2 x_2_blue x_2_green +2 x_2_blue x_2_red +2 x_2_green x_2_red +2 x_3_blue x_3_green " \
                             "+2 x_3_blue x_3_red +2 x_3_green x_3_red +2 x_4_blue x_4_green +2 x_4_blue x_4_red " \
                             "+2 x_4_green x_4_red +2 x_5_blue x_5_green +2 x_5_blue x_5_red +2 x_5_green x_5_red )\n" \
              "s.t. x_*_* in {0, 1}"
    assert str(objective_terms) == exp_str

    objective_terms[ONE_COLOR_PER_NODE] = objective_terms[ONE_COLOR_PER_NODE].to_ising() \
                                          .replace_variables(lambda tup: ("s", *tup[1:]))
                                          # change from "x" to "s" in the variable tuples
    exp_str = "min  P_0 * ( +1 x_1_blue x_2_blue +1 x_1_blue x_3_blue +1 x_1_blue x_5_blue +1 x_1_green x_2_green " \
                           "+1 x_1_green x_3_green +1 x_1_green x_5_green +1 x_1_red x_2_red +1 x_1_red x_3_red " \
                           "+1 x_1_red x_5_red +1 x_2_blue x_3_blue +1 x_2_green x_3_green +1 x_2_red x_3_red " \
                           "+1 x_3_blue x_4_blue +1 x_3_blue x_5_blue +1 x_3_green x_4_green +1 x_3_green x_5_green " \
                           "+1 x_3_red x_4_red +1 x_3_red x_5_red +1 x_4_blue x_5_blue +1 x_4_green x_5_green " \
                           "+1 x_4_red x_5_red )\n" \
              "     + P_1 * ( +5 +0.5 s_1_blue +0.5 s_1_green +0.5 s_1_red +0.5 s_2_blue +0.5 s_2_green +0.5 s_2_red " \
                             "+0.5 s_3_blue +0.5 s_3_green +0.5 s_3_red +0.5 s_4_blue +0.5 s_4_green +0.5 s_4_red " \
                             "+0.5 s_5_blue +0.5 s_5_green +0.5 s_5_red +0.5 s_1_blue s_1_green " \
                             "+0.5 s_1_blue s_1_red +0.5 s_1_green s_1_red +0.5 s_2_blue s_2_green " \
                             "+0.5 s_2_blue s_2_red +0.5 s_2_green s_2_red +0.5 s_3_blue s_3_green " \
                             "+0.5 s_3_blue s_3_red +0.5 s_3_green s_3_red +0.5 s_4_blue s_4_green " \
                             "+0.5 s_4_blue s_4_red +0.5 s_4_green s_4_red +0.5 s_5_blue s_5_green " \
                             "+0.5 s_5_blue s_5_red +0.5 s_5_green s_5_red )\n" \
              "s.t. s_*_* in {-1, 1},\n" \
              "     x_*_* in {0, 1}"
    assert str(objective_terms) == exp_str

def test_get_from_instance(objective_terms_direct_impl, exp_objective_terms, instance):
    """ test different initialization methods """
    # note that at the same time, this also tests two different implementations:
    #     - the fixture objective_terms relies on the implementation of the ExampleConstrainedObjective,
    #       from which we get then automatically generate the corresponding ObjectiveTerms
    #     - objective_terms_get uses the direct implementation of the ExampleObjectiveTerms
    assert objective_terms_direct_impl == exp_objective_terms

    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        objective_terms_depr = ExampleObjectiveTerms(instance=instance)
    assert objective_terms_depr == exp_objective_terms

def test_get_objective(objective_terms, exp_objective):
    """ test objective creation """
    terms_weights = {COLORED_EDGES : 1, ONE_COLOR_PER_NODE : 7}
    poly = objective_terms.get_weighted_sum_polynomial(terms_weights)
    assert isinstance(poly, PolyBinary)
    assert poly == exp_objective.polynomial

    objective = objective_terms.get_objective(terms_weights)
    assert objective == exp_objective

    objective = objective_terms.get_objective(penalty_scale=7)
    assert objective == exp_objective

    with pytest.warns(UserWarning, match="If the terms weights are given, the penalty scale is ignored"):
        objective_terms.get_objective(terms_weights, penalty_scale=7)
    assert objective == exp_objective

    with pytest.raises(ValueError, match="There is a mismatch between given and stored objective_terms names"):
        objective_terms.get_objective({"wrong_pw" : 4})
    with pytest.raises(ValueError, match="Cannot add PolyIsing and PolyBinary"):
        objective_terms.get_weighted_sum_polynomial({COLORED_EDGES : 1.0, ONE_COLOR_PER_NODE : PolyIsing()})

def test_evaluate_terms(objective_terms, solution):
    """ test check solution """
    values = objective_terms.evaluate_terms(solution)
    assert values[COLORED_EDGES] == 0.0
    assert values[ONE_COLOR_PER_NODE] == 0.0

def test_check_validity(objective_terms, solution):
    """ test check solution """
    assert objective_terms.check_validity(solution)

    solution[(X, 2, "green")] = 1
    assert not objective_terms.check_validity(solution)

def test_get_string(objective_terms):
    """ test if unique string is generated correctly """
    terms_weights = {COLORED_EDGES : 1.0, ONE_COLOR_PER_NODE : 2.4}
    exp_str = "colored_edges1.000000e+00_one_color_per_node2.400000e+00"
    assert objective_terms.get_terms_weights_string(terms_weights) == exp_str

def test_objective_terms_creation_no_impl():
    """ test objective terms creation """
    colored_edges = PolyBinary.read_from_string("5 + x0 x1 + x1 x2 + x2 x3 + x0 x3 + x0 x2")
    one_color_per_node = PolyBinary.read_from_string("1 - x0")
    objective_terms_dict = {"colored_edges": colored_edges, "one_color_per_node": one_color_per_node}
    constraint_terms_names = [ONE_COLOR_PER_NODE]
    objective_terms = ObjectiveTerms(objective_terms_dict, constraint_terms_names)
    assert objective_terms.constraint_terms_names == constraint_terms_names
    assert dict(objective_terms) == objective_terms_dict

    # explicitly setting constraint terms names to [] is fine
    objective_terms2 = ObjectiveTerms(objective_terms_dict, [])
    assert objective_terms2.constraint_terms_names == []
    assert dict(objective_terms2) == objective_terms_dict
    assert objective_terms2 != objective_terms

    with pytest.warns(UserWarning, match="There is nothing in this ObjectiveTerms"):
        ObjectiveTerms()

def test_errors(objective_terms):
    """ test the errors """
    # pylint: disable=protected-access
    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        ExampleObjectiveTerms(instance=ExampleInstance([(0, 1)], 1))
    objective_terms_dict = dict(objective_terms)
    with pytest.raises(NotImplementedError, match="Provide 'objective_terms' on initialization or "):
        ObjectiveTerms._get_objective_terms(None)
    with pytest.raises(NotImplementedError, match="Provide 'constraint_terms_names' on initialization or "):
        ObjectiveTerms._get_constraint_terms_names()
    with pytest.raises(ValueError, match="All constraint terms names should also be a key of the objective terms"):
        ObjectiveTerms(objective_terms_dict, ["wrong_constraint_name"])

def test_all_to_binary(objective_terms_different_types):
    """ test transformation of all terms to binary polynomials """
    objective_terms_binary = objective_terms_different_types.all_to_binary()
    exp_transformed_ising = PolyBinary({(): 6, (0,): -4, (0, 1): 8, (1,): -4, (2,): -8, (2, 3): 16, (3,): -8})
    assert isinstance(objective_terms_binary, ExampleObjectiveTerms)
    assert all(isinstance(poly, PolyBinary) for poly in objective_terms_binary.values())
    assert objective_terms_binary["binary"] == objective_terms_different_types["binary"]
    assert objective_terms_binary["none"] == objective_terms_different_types["none"]
    assert objective_terms_binary["ising"] == exp_transformed_ising

def test_all_to_ising(objective_terms_different_types):
    """ test conversion of all terms to Ising polynomials """
    # without inversion of ising
    objective_terms_ising = objective_terms_different_types.all_to_ising()
    assert isinstance(objective_terms_ising, ExampleObjectiveTerms)
    assert all(isinstance(poly, PolyIsing) for poly in objective_terms_ising.values())
    assert all(not poly.is_inverted() for poly in objective_terms_ising.values())

    # with inversion of ising
    objective_terms_ising = objective_terms_different_types.all_to_ising(inverted=True)
    assert isinstance(objective_terms_ising, ExampleObjectiveTerms)
    assert all(isinstance(poly, PolyIsing) for poly in objective_terms_ising.values())
    assert all(poly.is_inverted() for poly in objective_terms_ising.values())

def test_get_default_terms_weights(objective_terms):
    """ test the default weights """
    exp_weights = {"colored_edges": 1, "one_color_per_node": 1}
    assert objective_terms.get_default_terms_weights() == exp_weights

    exp_weights = {"colored_edges": 1, "one_color_per_node": 3}
    assert objective_terms.get_default_terms_weights(penalty_scale=3) == exp_weights

    exp_weights = {"colored_edges": 1, "one_color_per_node": 22}
    assert objective_terms.get_default_terms_weights(use_naive_bounds=True) == exp_weights

    exp_weights = {"colored_edges": 1, "one_color_per_node": 2}
    assert objective_terms.get_default_terms_weights(use_naive_bounds=True, penalty_scale=1/11) == exp_weights

def test_get_naive_terms_weights(objective_terms):
    """ test the naive weights """
    with pytest.raises(ValueError, match="Either provide ObjectiveTerms objects or the constraint terms names"):
        get_naive_terms_weights(dict(objective_terms))

    exp_naive_terms_weights = {"colored_edges": 1, "one_color_per_node": 22}
    assert get_naive_terms_weights(objective_terms) == exp_naive_terms_weights

    # check constraint for non-integer coefficients with factor 0.3
    constraint_poly = PolyBinary({((X, 1), (X, 2)): 1, ((X, 2),): -1.4, ((X, 1),): 5, ((X, 3),): 3})
    objective_terms["float_constraint"] = constraint_poly

    # the name of the constraint is not added to the constraint terms names, yet,
    # therefore is considered as objective and raises the error
    with pytest.raises(ValueError, match="Currently only one objective function is supported"):
        get_naive_terms_weights(objective_terms)
    objective_terms.constraint_terms_names.append("float_constraint")

    exp_naive_terms_weights = {"colored_edges": 1, "one_color_per_node": 22, "float_constraint":  55}
    assert get_naive_terms_weights(objective_terms) == exp_naive_terms_weights

    # remove objective function
    del objective_terms["colored_edges"]
    exp_naive_terms_weights = {"one_color_per_node": 1, "float_constraint": 1}
    assert get_naive_terms_weights(objective_terms) == exp_naive_terms_weights

def test_get_all_variables(objective_terms):
    """ test all variables """
    exp_variables = [(X, 1, B), (X, 1, G), (X, 1, R),
                     (X, 2, B), (X, 2, G), (X, 2, R),
                     (X, 3, B), (X, 3, G), (X, 3, R),
                     (X, 4, B), (X, 4, G), (X, 4, R),
                     (X, 5, B), (X, 5, G), (X, 5, R)]
    variables = objective_terms.get_all_variables()
    assert variables == exp_variables

def test_reduce():
    """ test reduction """
    term1 = PolyBinary({((X, 1), (X, 2), (X, 3)): 1, ((X, 1),) : -1})
    term2 = PolyBinary({((X, 2), (X, 3)): 2})
    objective_terms = ObjectiveTerms({"term1": term1, "term2": term2}, ["term1"])

    # has already only qubic polynomials
    reduced = objective_terms.reduce(3)
    assert reduced == objective_terms

    # only first term is reduced
    exp_reduction_term1 = PolyBinary({((REDUCTION, X, 1, X, 2),): 3,
                                      ((REDUCTION, X, 1, X, 2), (X, 1)): -2,
                                      ((REDUCTION, X, 1, X, 2), (X, 2)): -2,
                                      ((X, 1), (X, 2)): 1})
    exp_reduced_term1 = PolyBinary({((REDUCTION, X, 1, X, 2), (X, 3)): 1, ((X, 1),) : -1})
    exp_reduced = ObjectiveTerms({"term1": exp_reduced_term1, "term2": term2, "reduction_x_1_x_2": exp_reduction_term1},
                                 ["term1", "reduction_x_1_x_2"])
    reduced = objective_terms.reduce(single_penalty_term=False)
    assert reduced == exp_reduced

    # in the first step term1 is reduced, this is applied to term2, and afterwards term2 is further reduced
    term2 = PolyBinary({((X, 0), (X, 1), (X, 2), (X, 3)): 2})
    objective_terms = ObjectiveTerms({"term1": term1, "term2": term2}, ["term1"])

    exp_reduced_term2 = PolyBinary({((REDUCTION, REDUCTION, X, 1, X, 2, X, 0), (X, 3)): 2})
    exp_reduction_term2 = PolyBinary({((REDUCTION, REDUCTION, X, 1, X, 2, X, 0),): 3,
                                      ((REDUCTION, REDUCTION, X, 1, X, 2, X, 0), (X, 0)): -2,
                                      ((REDUCTION, REDUCTION, X, 1, X, 2, X, 0), (REDUCTION, X, 1, X, 2)): -2,
                                      ((REDUCTION, X, 1, X, 2), (X, 0)): 1})
    exp_reduction = exp_reduction_term1 + exp_reduction_term2
    exp_reduced = ObjectiveTerms({"term1": exp_reduced_term1, "term2": exp_reduced_term2, REDUCTION: exp_reduction},
                                 ["term1", REDUCTION])
    reduced = objective_terms.reduce()
    assert reduced == exp_reduced

    # check errors
    with pytest.raises(ValueError, match="Can only reduce to at least quadratic"):
        ObjectiveTerms({"term1": term1}, []).reduce(1)
    with pytest.raises(ValueError, match="Reduction is currently only implemented for binary polynomials"):
        ObjectiveTerms({"term1": PolyIsing()}, []).reduce()



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
module for decorators for adding IO functions to classes

thanks to https://medium.com/@mgarod/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
"""

from functools import wraps

from . import hdf5


def self_method(cls, method_name=None):
    """
    get a decorator which adds the decorated function to the given class as a self method,
    for an object of the given class (obj = cls(...)) instead of calling the function with 'decorated_func(obj, ...)',
    the following method is available AT RUNTIME 'obj.decorated_func(...)' (if IO is loaded),
    by adding a 'method_name' the functionality is preserved, but the call changes to 'obj.method_name(...)',
    can be used to 'register' the class to provide a certain functionality

    :param cls: the class to add the method to
    :param (str or None) method_name: the name of the added method, by default name of the decorated function is used
    :return: the decorator
    """
    def decorator(decorated_func):
        """ the actual decorator """

        @wraps(decorated_func)  # preserve name and docstring
        def wrapper(self, *args, **kwargs):
            """ wrapper accepts self and inserts it as the first argument of the decorated function """
            return decorated_func(self, *args, **kwargs)

        # if no explicit name is given use the one of the decorated function
        name = method_name or decorated_func.__name__
        # add the function as a method to the class
        setattr(cls, name, wrapper)
        # with returning the decorated function, it can still be used normally
        return decorated_func

    return decorator


def class_method(cls, method_name=None):
    """
    get a decorator which adds the decorated function to the given class as a class method,
    for the given class instead of calling the function with 'decorated_func(cls, ...)',
    the following method is available AT RUNTIME 'cls.decorated_func(...)' (if IO is loaded),
    by adding a 'method_name' the functionality is preserved, but the call changes to 'cls.method_name(...)',
    can be used to 'register' the class to provide a certain functionality

    :param cls: the class to add the method to
    :param (str or None) method_name: the name of the added method, by default name of the decorated function is used
    :return: the decorator
    """
    def decorator(decorated_func):
        """ the actual decorator """

        @classmethod
        @wraps(decorated_func)  # preserve name and docstring
        def wrapper(*args, **kwargs):
            """ wrapper does not need cls or self """
            return decorated_func(*args, **kwargs)

        # if no explicit name is given use the one of the decorated function
        name = method_name or decorated_func.__name__
        # add the function as a method to the class
        setattr(cls, name, wrapper)
        # with returning the decorated function, it can still be used normally
        return decorated_func

    return decorator

def static_method(cls, method_name=None):
    """
    get a decorator which adds the decorated function to the given class as a static method,
    for the given class instead of calling the function with 'decorated_func(...)',
    the following method is available AT RUNTIME 'cls.decorated_func(...)' (if IO is loaded),
    by adding a 'method_name' the functionality is preserved, but the call changes to 'cls.method_name(...)',
    can be used to 'register' the class to provide a certain functionality

    :param cls: the class to add the method to
    :param (str or None) method_name: the name of the added method, by default name of the decorated function is used
    :return: the decorator
    """
    def decorator(decorated_func):
        """ the actual decorator """

        @staticmethod
        @wraps(decorated_func)
        def wrapper(*args, **kwargs):
            """ wrapper does not need cls or self """
            return decorated_func(*args, **kwargs)

        # if no explicit name is given use the one of the decorated function
        name = method_name or decorated_func.__name__
        # add the function as a method to the class
        setattr(cls, name, wrapper)
        # with returning the decorated function, it can still be used normally
        return decorated_func

    return decorator

def save_load_exists(the_class, module=hdf5, suffix="hdf5"):
    """
    add all IO methods to the class definition,
    those call e.g. hdf5.save, hdf5.load and hdf5.exists, respectively

    :param the_class: the class to which the methods shall be added
    :param module: the module from which the methods are taken
    :param suffix: the suffix to add to the method name
    """
    self_method(the_class, f"save_{suffix}")(module.save)
    class_method(the_class, f"load_{suffix}")(module.load)
    class_method(the_class, f"exists_{suffix}")(module.exists)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ExampleConstraintObjective """

from quark import ConstrainedObjective, ConstraintBinary, PolyBinary


X = "x"
ONE_COLOR_PER_NODE = "one_color_per_node"
COLORED_EDGES      = "colored_edges"


class ExampleConstrainedObjective(ConstrainedObjective):
    """
    An example implementation of a ConstraintObjective for testing based on the MaxColorableSubgraph problem

    The binary variable x_n_c is 1 if node n is colored with color c and 0 otherwise.
    """

    @staticmethod
    def _get_objective_poly(instance):
        """ get the objective polynomial from the instance data """
        # counting the number of wrongly colored edges:
        # sum_[c in Colors] sum_[(n, m) in Edges] (1 * x_n_c * x_m_c)
        return PolyBinary({((X, node1, color), (X, node2, color)): 1 for node1, node2 in instance.edges
                                                                     for color in instance.colors})

    @staticmethod
    def _get_constraints(instance):
        """ get the constraints from the instance data """
        constraints = {}

        # every node should get exactly one color:
        # for all n in Nodes: sum_[c in Colors] x_n_c == 1
        for node in instance.nodes:
            poly = PolyBinary({((X, node, color),): 1 for color in instance.colors})
            constraints[ONE_COLOR_PER_NODE + f"_{node}"] = ConstraintBinary(poly, 1, 1)

        return constraints

    def get_objective_terms(self, objective_name=COLORED_EDGES, combine_prefixes=(ONE_COLOR_PER_NODE,), name=None,
                            reduction_strategy=None, check_special_constraints=True):
        """
        get an ObjectiveTerms object with the objective as one term
        and automatically generated penalty terms for each constraint

        this overwrites the inherited method to set specific default values
        """
        return super().get_objective_terms(objective_name=objective_name, combine_prefixes=combine_prefixes, name=name,
                                           reduction_strategy=reduction_strategy,
                                           check_special_constraints=check_special_constraints)

    @staticmethod
    def get_original_problem_solution(raw_solution, precision=1e-9):
        """
        extract the actual solution from the variable assignment

        :param (Solution or dict) raw_solution: the binary assignment of the variables according to the above encoding
        :param (float) precision: precision value for rounding issues
        :return: the actual mapping of interest from nodes to colors
        """
        return {node : color for (_, node, color), value in raw_solution.items() if abs(value - 1) < precision}



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the ObjectiveTerms """

import os
import pytest

from quark import PolyBinary, ObjectiveTerms
from quark.io import hdf5
from quark.testing import ExampleObjectiveTerms


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "objective_terms.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE


def test_load_hdf5(objective_terms, objective_terms_direct_impl):
    """ test hdf5 io """
    loaded = hdf5.load(ObjectiveTerms, FILENAME_FIX)
    assert loaded == objective_terms

    loaded = hdf5.load(ExampleObjectiveTerms, FILENAME_FIX)
    assert loaded == objective_terms_direct_impl

def test_io_hdf5(objective_terms, objective_terms_direct_impl):
    """ test hdf5 io round trip """
    hdf5.save(objective_terms, FILENAME_TMP)
    loaded = hdf5.load(ObjectiveTerms, FILENAME_TMP)
    assert loaded == objective_terms

    hdf5.save(objective_terms_direct_impl, FILENAME_TMP)
    loaded = hdf5.load(ExampleObjectiveTerms, FILENAME_TMP)
    assert loaded == objective_terms

    os.remove(FILENAME_TMP)


@pytest.fixture
def objective_terms_no_impl():
    """ provide the ObjectiveTerms test object directly using the base class """
    objective = PolyBinary.read_from_string("5 + x0 x1 + x1 x2 + x2 x3 + x0 x3 + x0 x2")
    constraint = PolyBinary.read_from_string("1 - x0")
    ot_dict = dict(objective=objective, constraint=constraint)
    objective_terms = ObjectiveTerms(ot_dict, ["constraint"], name="test_objective_terms")
    yield objective_terms

def test_load_hdf5_no_impl(objective_terms_no_impl):
    """ test hdf5 io """
    loaded = hdf5.load(ObjectiveTerms, FILENAME_FIX, name="test_objective_terms", prefix_group_name="elsewhere")
    assert loaded == objective_terms_no_impl

def test_io_hdf5_no_impl(objective_terms_no_impl):
    """ test hdf5 io round trip """
    hdf5.save(objective_terms_no_impl, FILENAME_TMP, prefix_group_name="elsewhere")
    loaded = hdf5.load(ObjectiveTerms, FILENAME_TMP, name="test_objective_terms", prefix_group_name="elsewhere")
    assert loaded == objective_terms_no_impl

    os.remove(FILENAME_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Testing objective.py """

import pytest

from quark import VariableMapping, Polynomial, PolyIsing, PolyBinary, Objective
from quark.testing import ExampleInstance, ExampleObjective


X, Y = "x", "y"
NAME = "test_objective"
CONSTRUCTED_NAME = 'colored_edges1.000000e+00_one_color_per_node7.000000e+00'

POLY = Polynomial.read_from_string("5 x1 + 6 x2 + 3 x1 x3 - x4 - 3 x2 x4 + 3")
BINARY = PolyBinary.read_from_string("5 x1 + 6 x2 + 3 x1 x3 - x4 - 3 x2 x4 + 3")
ISING = 0.25 * PolyIsing.read_from_string("32 + 13 x1 + 9 x2 + 3 x3 - 5 x4 + 3 x1 x3 - 3 x2 x4")
ISING_INV = 0.25 * PolyIsing(Polynomial.read_from_string("32-13 x1 -9 x2 -3 x3 +5 x4 +3 x1 x3 -3 x2 x4"), inverted=True)
SYM_ISING = PolyIsing({(0, 1): 0.5, (0, 5): 0.5, (1, 2): 0.5, (2, 3): 0.5, (3, 4): 0.5, (4, 5): 0.5}, inverted=False)

OBJ_UNDEF = Objective(POLY, NAME)
OBJ_BINARY = Objective(BINARY, NAME)
OBJ_ISING = Objective(ISING, NAME)
OBJ_ISING_INV = Objective(ISING_INV, NAME)
OBJ_SYM_ISING = Objective(SYM_ISING, NAME)


def test_init(objective, exp_objective):
    """ test objective creation """
    assert objective == exp_objective

def test_get_from_instance(objective_direct_impl, exp_objective, instance):
    """ test different initialization methods """
    # note that at the same time, this also tests two different implementations:
    #     - the fixture objective relies on the implementation of the ExampleConstrainedObjective,
    #       from which we get then automatically generate the corresponding ObjectiveTerms and the Objective
    #     - objective_get uses the direct implementation of the ExampleObjective
    assert not objective_direct_impl.name
    objective_direct_impl.name = CONSTRUCTED_NAME
    assert objective_direct_impl == exp_objective

    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        objective_depr = ExampleObjective(instance=instance, name=CONSTRUCTED_NAME)
    assert objective_depr == exp_objective

def test_repr():
    """ test string representation """
    OBJ_BINARY.name = "test_objective"
    exp_str = "Objective(PolyBinary({(): 3.0, (1,): 5.0, (2,): 6.0, (4,): -1.0, (1, 3): 3.0, (2, 4): -3.0}), " \
                        "'test_objective')"
    assert repr(OBJ_BINARY) == exp_str

def test_str():
    """ test string representation """
    exp_str = "min  +3 +5 x1 +6 x2 -1 x4 +3 x1 x3 -3 x2 x4\n" \
              "s.t. x* in {0, 1}"
    assert str(OBJ_BINARY) == exp_str

    poly = PolyBinary({(): 3, ((X, 1),): 5, ((X, 2),): 6, ((Y, 4, 1),): -1,
                       ((X, 1), (Y, 3, 1)): 3, ((X, 2), (Y, 4, 1)): -3})
    exp_str = "min  +3 +5 x_1 +6 x_2 -1 y_4_1 +3 x_1 y_3_1 -3 x_2 y_4_1\n" \
              "s.t. x_* in {0, 1}\n" \
              "     y_*_* in {0, 1}"
    assert str(Objective(poly)) == exp_str

    exp_str = "min  +8 -3.25 s1 -2.25 s2 -0.75 s3 +1.25 s4 +0.75 s1 s3 -0.75 s2 s4\n" \
              "s.t. s* in {-1, 1}"
    assert str(OBJ_ISING_INV) == exp_str

def test_is_binary():
    """ test is_binary method """
    assert OBJ_BINARY.is_binary()
    assert not OBJ_ISING.is_binary()
    assert not OBJ_ISING_INV.is_binary()
    assert not OBJ_UNDEF.is_binary()

def test_is_ising():
    """ test is_ising method """
    # ising without inverted flag
    assert OBJ_ISING.is_ising()
    assert OBJ_ISING_INV.is_ising()
    assert not OBJ_BINARY.is_ising()
    assert not OBJ_UNDEF.is_ising()

    # non-inverted ising
    assert OBJ_ISING.is_ising(inverted=False)
    assert not OBJ_ISING_INV.is_ising(inverted=False)
    assert not OBJ_BINARY.is_ising(inverted=False)
    assert not OBJ_UNDEF.is_ising(inverted=False)

    # inverted ising
    assert not OBJ_ISING.is_ising(inverted=True)
    assert OBJ_ISING_INV.is_ising(inverted=True)
    assert not OBJ_BINARY.is_ising(inverted=True)
    assert not OBJ_UNDEF.is_ising(inverted=True)

def test_to_ising():
    """ test conversion to ising """
    # binary to ising
    assert OBJ_BINARY.to_ising() == OBJ_ISING
    assert OBJ_BINARY.to_ising() != ISING
    assert OBJ_BINARY.to_ising(invert=True) == OBJ_ISING_INV

    # should be identical
    assert OBJ_ISING.to_ising() == OBJ_ISING
    assert OBJ_ISING_INV.to_ising() == OBJ_ISING_INV

    # inversion
    assert OBJ_ISING_INV.to_ising(invert=True) == OBJ_ISING
    assert OBJ_ISING.to_ising(invert=True) == OBJ_ISING_INV

def test_to_binary():
    """ test conversion to binary """
    assert OBJ_ISING.to_binary() == OBJ_BINARY
    assert OBJ_ISING_INV.to_binary() == OBJ_BINARY
    assert OBJ_UNDEF.to_binary() == OBJ_BINARY
    assert OBJ_BINARY.to_binary() == OBJ_BINARY

def test_compact():
    """ test making the polynomial compact """
    compact, var_mapping = OBJ_BINARY.compact()
    exp_poly = PolyBinary(Polynomial.read_from_string("5 x0 + 6 x1 + 3 x0 x2 - x3 - 3 x1 x3 + 3"))
    exp_vm = VariableMapping({0: 1, 1: 2, 2: 3, 3: 4})
    assert compact.polynomial == exp_poly
    assert compact.name == NAME
    assert var_mapping == exp_vm

    compact, var_mapping = OBJ_ISING_INV.compact(new_name="new_name")
    exp_poly = 0.25 * PolyIsing(Polynomial.read_from_string("32 - 13 x0 - 9 x1 - 3 x2 + 5 x3 + 3 x0 x2 - 3 x1 x3"),
                                inverted=True)
    assert compact.polynomial == exp_poly
    assert compact.name == "new_name"
    assert var_mapping == exp_vm

    compact = Objective(PolyBinary({(0,): 1, (1, 0): 2}), "compact")
    assert compact.compact() == (compact, None)

def test_break_symmetry_by_fixing_variable():
    """ test fixing one variable """
    exp_name = "objective_fixed_spin"
    exp_poly = PolyIsing({(0, 1): 0.5, (0,): 0.5, (1, 2): 0.5, (2, 3): 0.5, (3, 4): 0.5, (4,): 0.5})
    new_objective = OBJ_SYM_ISING.break_symmetry_by_fixing_variable(new_name=exp_name)
    assert new_objective.polynomial == exp_poly
    assert new_objective.name == exp_name

    with pytest.raises(ValueError, match="Only supported for Ising models"):
        OBJ_BINARY.break_symmetry_by_fixing_variable()
    with pytest.raises(ValueError, match="Only supported for Ising models without linear terms"):
        OBJ_ISING.break_symmetry_by_fixing_variable()
    with pytest.raises(ValueError, match="Cannot find variable '6' in polynomial"):
        OBJ_SYM_ISING.break_symmetry_by_fixing_variable(new_name=exp_name, variable=6)

def test_errors():
    """ test the errors """
    # pylint: disable=protected-access
    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        ExampleObjective(instance=ExampleInstance([(0, 1)], 1))
    with pytest.raises(NotImplementedError, match="Provide 'polynomial' on initialization or "):
        Objective._get_polynomial(None)
    with pytest.raises(ValueError, match="Objective's polynomial is neither Polynomial, PolyIsing nor PolyBinary"):
        Objective("13 x1 + 15 x2", "wrong")
    with pytest.warns(UserWarning, match="There is nothing in this Objective"):
        Objective(PolyBinary())



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for IO of attributes with HDF5 files """

import collections.abc
from numbers import Integral, Real
import numpy as np


ERROR_DATASET   = "Given value does not have an easy data structure, create dataset instead"
ERROR_NOT_FOUND = "Did not find attribute '{}' at group '{}' in HDF5 file '{}'"


def write_attribute(group, name, obj=None, value=None):
    """
    write attribute at the open group of an HDF5 file,
    either the value of the attribute is taken from the provided object or needs to be given explicitly

    :param (h5py.Group) group: the group in the HDF5 file at which the attribute will be attached
    :param (str) name: the name under which the attribute is stored
    :param obj: the object with an attribute called as name, whose value is taken as the attribute to be stored
    :param value: the explicit attribute value which shall be stored if no object is given
    """
    # if the object is given it should have an attribute which is called name
    if obj is not None and hasattr(obj, name):
        value = getattr(obj, name)
    value = _format_attribute(value)

    # if the attribute already exists at the group it will be deleted and overwritten
    if name in group.attrs:
        del group.attrs[name]
    group.attrs.create(name, value)

def write_attributes(group, obj=None, *names, **names_to_values):  # pylint: disable=keyword-arg-before-vararg
    """
    write several attributes at the open group of an HDF5 file,
    either provide the names of the attributes of the given object yielding the values or
    give the explicit dictionary of names to values as keyword arguments

    :param (h5py.Group) group: the HDF5 group to store the attributes at
    :param obj: the object whose attributes to take as attributes to be stored
    :param (str) names: attributes of the object that shall be stored (not to be used if no object is given)
    :param names_to_values: keyword arguments of explicit attribute values with names,
                            if no object is given or additional values shall be stored
    """
    for name in names:
        write_attribute(group, name, obj)
    for name, value in names_to_values.items():
        write_attribute(group, name, value=value)

def _format_attribute(value):
    """
    check if value has correct type and reformat it if necessary

    :param value: the attribute value
    :return the corrected value
    """
    if value is None:
        return np.NaN
    if isinstance(value, str):
        return np.string_(value)
    if not np.isscalar(value):
        if isinstance(value, set):
            value = list(value)
        if not isinstance(value, collections.abc.Sequence) or not all(np.isscalar(v) for v in value):
            raise ValueError(ERROR_DATASET)
    return value

def read_attribute(group, name, check_existence=True):
    """
    read attribute from the open group of an HDF5 file

    :param (h5py.Group) group: the group in the HDF5 file where the attribute is stored
    :param (str) name: the name under which the attribute is stored
    :param (bool) check_existence: if True then an error will be thrown if the attribute is not found,
                                   if False None will be returned without error
    :return: the attribute value
    """
    # if it is there and not Nan the found value will be returned
    if name in group.attrs:
        value = group.attrs[name]
        if not (isinstance(value, float) and np.isnan(value)):
            if isinstance(value, bytes):
                return value.decode()
            if isinstance(value, Integral):
                return int(value)
            if isinstance(value, Real):
                return float(value)
            return value

    # if is not found and the existence is obligatory
    elif check_existence:
        raise ValueError(ERROR_NOT_FOUND.format(name, group.name, group.file.filename))

    # if it does not exist or is Nan
    return None

def read_attributes(group, *names, check_existence=True):
    """
    read several attributes from the open group of an HDF5 file

    :param (h5py.Group) group: the group in the HDF5 file where the attributes are stored
    :param (str) names: the names under which the attributes are stored
    :param (bool) check_existence: if True then an error will be thrown if an attribute is not found,
                                   if False None will be returned without error
    :return: dictionary with the attributes keyed by names
    """
    return {name : read_attribute(group, name, check_existence) for name in names}



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for PolyBinary """

from collections import Counter
from functools import cached_property
from warnings import warn
from itertools import combinations

from .polynomial import Polynomial, sort_poly_dict
from . import poly_ising
from .utils.variables import to_string


ADD       = "add"
MULTIPLY  = "multiply"
REDUCTION = "reduction"
X         = "x"

ERROR_ISING      = "Cannot {} PolyBinary and PolyIsing"
ERROR_DEGREE     = "Not applicable for polynomials with degree larger than 2"
ERROR_UPPER      = "Upper bound needs to be larger than or equal to lower bound"
ERROR_INPUT      = "Either provide upper_bound or num_variables"
ERROR_MAX_DEGREE = "Provide integer max_degree larger than 0 or set to None to not use automatic reduction"

WARNING_VARS   = "Variables are replaced by tuple variables"
WARNING_IGNORE = "With upper_bound given, num_variables is ignored"

STRAIGHT = "straight"
SIMPLE   = "simple"
MEDIUM   = "medium"
BETTER   = "better"


class PolyBinary(Polynomial):
    """
    A binary polynomial is a special polynomial, where the variables can only take 0 or 1.
    Since 0^2 = 0 and 1^2 = 1, we can simplify the Polynomial by removing all exponents,
    e.g.

        * (0, 0, 1, 1, 1): 1 -> (0, 1): 1,
        * (2, 2, 2, 2): 2    -> (2,): 2.

    We have the following conventions:

        *  PolyBinary == PolyBinary <=> poly == poly,
        *  PolyBinary == Polynomial <=> poly == poly,
        *  PolyBinary +/* Polynomial = PolyBinary,
        *  Polynomial +/* PolyBinary = Polynomial + Warning,
        *  PolyIsing +/* PolyBinary = Error,
        *  copy/remove_zero_coefficients/replace_variables/get_rounded(PolyBinary) = PolyBinary,
        *  evaluate(PolyBinary, {var : Number}) = PolyBinary/Number,
        *  evaluate(PolyBinary, {var : Polynomial}) = Polynomial,
        *  evaluate(PolyBinary, {var : PolyBinary}) = Error,
        *  PolyBinary == PolyIsing = False.
    """

    @cached_property
    def positive(self):
        """ part of the polynomial with positive coefficients without offset """
        return self.__class__({var_tuple: coeff for var_tuple, coeff in self.items() if coeff > 0 and var_tuple},
                              variable_tuples_already_formatted=True)

    @cached_property
    def negative(self):
        """ part of the polynomial with negative coefficients without offset """
        return self.__class__({var_tuple: coeff for var_tuple, coeff in self.items() if coeff < 0 and var_tuple},
                              variable_tuples_already_formatted=True)

    @cached_property
    def naive_upper_bound(self):
        """
        a naive upper bound derived from the sum of all positive coefficients,
        it is a worst-case estimation and in general only tight, if the polynomial is linear,
        as it does not exploit any deeper structure in the polynomial
        """
        return sum(self.positive.values()) + self.offset

    @cached_property
    def naive_lower_bound(self):
        """
        returns a naive lower bound derived from the sum of all negative coefficients,
        it is a worst-case estimation and in general only tight, if the polynomial is linear,
        as it does not exploit any deeper structure in the polynomial
        """
        return sum(self.negative.values()) + self.offset

    def _format_variable_tuple(self, var_tuple):
        """
        transform the tuple of variables into the standard format, also checks the type of the variables,
        extends 'Polynomial._format_variable_tuple',
        since we have x * x = x  for all x in {0, 1}, all powers can be removed
        """
        var_tuple = super()._format_variable_tuple(var_tuple)
        var_tuple = tuple(sorted(set(var_tuple)))
        return var_tuple

    def __eq__(self, other):
        """
        check equality of polynomials, where PolyIsing is never equal to a PolyBinary,
        extends 'Polynomial.__eq__'
        """
        if isinstance(other, poly_ising.PolyIsing):
            return False
        return super().__eq__(other)

    @staticmethod
    def _check_different(poly2, func_name):
        """ check if the type of the second Polynomial is feasible for an operation """
        if isinstance(poly2, poly_ising.PolyIsing):
            raise ValueError(ERROR_ISING.format(func_name))

    def __add__(self, poly2):
        """
        add another polynomial or a scalar,
        extends 'Polynomial.__add__' since PolyIsing cannot be added to PolyBinary
        """
        self._check_different(poly2, ADD)
        return super().__add__(poly2)

    def __mul__(self, poly2):
        """
        multiply with another polynomial or a scalar,
        extends 'Polynomial.__mul__' since PolyIsing cannot be multiplied with PolyBinary
        """
        self._check_different(poly2, MULTIPLY)
        return super().__mul__(poly2)

    @classmethod
    def _is_valid_var_assignment(cls, var_assignment):
        """
        check whether the given input can be assigned to the variable,
        that means it is a Polynomial, PolyBinary, SCIP Expr or valid coefficient,
        extends 'Polynomial._is_valid_var_assignment' to not allow for PolyIsing
        """
        return not isinstance(var_assignment, PolyBinary) and super()._is_valid_var_assignment(var_assignment)

    def is_non_negative(self):
        """
        check if the polynomial has only non-negative coefficients
        and thus always evaluates to non-negative values
        """
        return all(coeff >= 0 for coeff in self.values())

    def get_matrix_representation(self, num_variables=None):
        """
        get the polynomial in the standard matrix representation

        :return: the quadratic matrix of the standard representation
        """
        linear, quadratic = super().get_matrix_representation(num_variables)
        for i, value in enumerate(linear):
            quadratic[i, i] = value
        return quadratic

    # additional methods for binary polynomials

    @classmethod
    def from_unknown_poly(cls, poly):
        """
        convert a polynomial of unknown type into a PolyBinary

        :param (Polynomial) poly: the polynomial of unknown type
        :return: the corresponding binary polynomial
        """
        return poly.to_binary() if isinstance(poly, poly_ising.PolyIsing) else cls(poly)

    def preprocess_minimize(self, unambitious=False):
        """
        simplify the polynomial by setting some variables to the values which are guaranteed to appear in an optimal
        solution when considering minimization

        :param (bool) unambitious:
            if True, only those variables will be set where their assignment is unambitious, that means, in all optimal
            solutions, they have the same value,
            if False, preprocess the variable to the straightforward value whether it might also get another value in an
            optimal solution, the objective values remains the same
        :return: the new reduced Polynomial and the assignment of the eliminated variables with variable to value
        """
        return super().preprocess(preprocess_rule, unambitious)

    def to_ising(self, inverted=False):
        """
        convert the binary polynomial to the Ising format

        :param (bool) inverted: if False, use conversion 0 -> -1, 1 -> 1 with x = 0.5s + 0.5,
                                if True, use conversion  0 -> 1, 1 -> -1 with x = -0.5s + 0.5
        :return: the corresponding PolyIsing
        """
        converted = self.affine_transform(-0.5, 0.5) if inverted else self.affine_transform(0.5, 0.5)
        return poly_ising.PolyIsing(converted, inverted)

    def reduce(self, max_degree=1, use=None, reduction_variable_prefix=(REDUCTION,), reduction_strategy=None):
        """
        reduce the polynomial by substituting every variable product with a new variable until it is quadratic
        and get the corresponding penalty terms enforcing the reduction

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (list[tuples] or set[tuples] or None) use: the predefined reductions which shall be used,
                                                          set with tuples in the format (var1, var2, reduction_var)
        :param (tuple or str) reduction_variable_prefix: prefix for the additional reduction variables that are created
        :param (str or None) reduction_strategy: the strategy how to choose the next variable pair
        :return: the reduced polynomial, the dictionary of penalty terms enforcing the reductions
        """
        if max_degree is not None and (not isinstance(max_degree, int) or max_degree < 1):
            raise ValueError(ERROR_MAX_DEGREE)

        max_degree = max_degree or self.degree
        poly = _apply_reduction(self, use)  # try here already in case the reductions are on flat variables
        if poly.degree <= max_degree:
            return poly, []

        if self.is_flat():
            # in the following we need to introduce new variables which are of type tuple
            # as we cannot mix up different types of variables we need to rename the flat variables
            warn(WARNING_VARS)
            poly = poly.replace_variables(lambda i: (X, i))
        elif self.var_type == str:
            warn(WARNING_VARS)
            poly = poly.replace_variables(lambda s: (s,))
        if not isinstance(reduction_variable_prefix, tuple):
            reduction_variable_prefix = (reduction_variable_prefix,)

        # try again in case the reductions are on non-flat variables
        poly = _apply_reduction(poly, use)
        return _reduce_poly_successively(poly, max_degree, reduction_variable_prefix, reduction_strategy)


def preprocess_rule(var_sigmas, coeff, unambitious=False):
    """
    very simple preprocessing function:
    if the weight of the variable is positive and extends the influence of incoming negative couplings,
    the variable can be assigned to 0,
    if the weight is negative and extends the influence of incoming positive couplings, can be assigned to 1

    :param (tuple) var_sigmas: sigma values (combined incoming strengths) of a variable of the polynomial
    :param (Real) coeff: coefficient on the linear term of the variable
    :param (bool) unambitious: if True, only those variables will be set where their assignment is unambitious,
                               that means, in all optimal solutions, they have the same value,
                               if False, preprocess the variable to the straightforward value whether it might also
                               get another value in an optimal solution, the objective values remains the same
    :return: the solution value or None if the variable cannot be preprocessed
    """
    if unambitious and coeff == -var_sigmas[1]:
        return None
    if coeff >= -var_sigmas[1]:
        return 0
    if coeff <= -var_sigmas[0]:
        return 1
    return None


def _apply_reduction(poly, reductions):
    if not reductions:
        return poly
    reductions = [reduction for reduction in reductions if len(reduction) == 3]
    change = True
    poly_dict = poly
    while change:
        # we need to have several iterations as reductions might be based upon each other
        for reduction_var, var1, var2 in reductions:
            poly_dict, change = _replace_variable_pair(poly_dict, var1, var2, reduction_var, return_change=True)
    return PolyBinary(poly_dict)

def _reduce_poly_successively(poly_dict, max_degree, reduction_variable_prefix, reduction_strategy):
    reductions = []
    poly_dict = sort_poly_dict(poly_dict, reverse_monomial_degree_order=True)  # sorted by decreasing monomial degrees
    current_degree = len(next(iter(poly_dict.keys())))

    while current_degree > max_degree:
        reduction_strategy = SIMPLE if current_degree <= 2 else (reduction_strategy or SIMPLE)
        var1, var2 = _get_best_variable_pair(poly_dict, current_degree, max_degree, strategy=reduction_strategy)
        reduction_var = reduction_variable_prefix + var1 + var2
        poly_dict = _replace_variable_pair(poly_dict, var1, var2, reduction_var)
        reductions.append((reduction_var, var1, var2))
        current_degree = max(map(len, poly_dict.keys()))
    return PolyBinary(poly_dict), sorted(set(reductions))

def _get_best_variable_pair(poly_dict, degree, max_degree, strategy=BETTER):
    if strategy == STRAIGHT:
        # simply get the first variable pair in the first monomial with degree larger than max_degree
        monomials = iter(poly_dict.keys())
        monomial = next(monomials)
        while len(monomial) <= max_degree:
            value = poly_dict.pop(monomial)
            poly_dict[monomial] = value  # append the monomial at the end of the poly dict again
            monomial = next(monomials)
        var1, var2 = monomial[0:2]
        return var1, var2

    if strategy == SIMPLE:
        # get the largest monomial (or one of them) and from it the first variable pair
        max_monomial = max(poly_dict.keys(), key=len)
        var1, var2 = max_monomial[0:2]
        return var1, var2

    if strategy == MEDIUM:
        # get the variable pair which appears in the most monomials of the largest degree
        max_var_pairs = sum((list(combinations(var_tuple, 2))
                            for var_tuple in poly_dict.keys() if len(var_tuple) == degree), start=[])
        return Counter(max_var_pairs).most_common(1)[0][0]

    if strategy == BETTER:
        # get the variable pair which appears most in all monomials
        all_var_pairs = sum((list(combinations(var_tuple, 2)) for var_tuple in poly_dict.keys()), start=[])
        return Counter(all_var_pairs).most_common(1)[0][0]

    raise ValueError(f"Unknown strategy '{strategy}'")

def _replace_variable_pair(poly_dict, var1, var2, reduction_var, return_change=False):
    reduced_poly_dict = {}
    change = False
    for var_tuple, coeff in poly_dict.items():
        if var1 in var_tuple and var2 in var_tuple:
            # the new variable tuple has the reduction variable at the end, which prefers non-reduction variables later,
            # this is kept until the polynomial is instantiated, then the variable tuples might be resorted
            new_var_tuple = tuple(var for var in var_tuple if var not in [var1, var2]) + (reduction_var,)
            reduced_poly_dict[new_var_tuple] = reduced_poly_dict.get(new_var_tuple, 0) + coeff
            change = True
        else:
            reduced_poly_dict[var_tuple] = reduced_poly_dict.get(var_tuple, 0) + coeff
    return reduced_poly_dict if not return_change else (reduced_poly_dict, change)


REDUCTION_PENALTY_POLY_FULL = PolyBinary({(0,): 3, (0, 1): -2, (0, 2): -2, (1, 2): 1}) # +3 x0 -2 x0 x1 -2 x0 x2 + x1 x2

def get_reduction_penalty_polynomial(product_xy, factor_x, factor_y):
    """
    get the quadratic polynomial that enforces a variable to be equal to the product of two variables
    (the variables need to be of the same type, thus either tuple, int or str)

    :param (tuple or int or str) product_xy: new binary variable that substitutes x*y
    :param (tuple or int or str) factor_x: first binary variable
    :param (tuple or int or str) factor_y: second binary variable
    :return: the PolyBinary that is zero if x * y = xy and positive otherwise
    """
    return REDUCTION_PENALTY_POLY_FULL.replace_variables([product_xy, factor_x, factor_y])

def get_all_reduction_penalty_terms(reductions):
    """
    get the corresponding penalty terms to the reductions

    :param (list[tuple]) reductions: list of 3-tuples with reduction variables
    :return: the penalty terms to enforce the reductions
    """
    penalty_terms = {}
    for reduction_var, var1, var2 in reductions:
        reduction_name = to_string(reduction_var)
        penalty_terms[reduction_name] = get_reduction_penalty_polynomial(reduction_var, var1, var2)
    return penalty_terms


def get_binary_representation_polynomial(upper_bound=None, lower_bound=0, num_variables=None, variable_prefix=(X,),
                                         split_signs=False):
    """
    get the linear polynomial representing an integer variable bounded between the lower bound and the upper bound
    using the binary representation with binary variables,
    the polynomial will be empty if upper_bound == lower_bound == 0,
    by providing the number of desired variables instead, the polynomial will represent integer numbers up to 2**n-1

    :param (Integer) upper_bound: value of the upper bound, can be 0 but not negative unless the lower bound is also set
    :param (Integer) lower_bound: value of the lower bound
    :param (Integer) num_variables: the number of binary variables to be used
    :param (tuple or str) variable_prefix: prefix for the binary variables that are created
    :param (bool) split_signs: if the lower bound is negative and the upper bound is positive,
                               the negative part and the positive part will be treated separately,
                               this leads to more variables but no constant and lower absolute coefficients,
                               note that this also means the representation of a value is not unique anymore
    :return: the linear polynomial representing the integer variable
    """
    if upper_bound is None and not num_variables:
        raise ValueError(ERROR_INPUT)
    if upper_bound is not None and num_variables:
        warn(WARNING_IGNORE)
        num_variables = None
    if upper_bound and upper_bound < lower_bound:
        raise ValueError(ERROR_UPPER)

    if not isinstance(variable_prefix, tuple):
        variable_prefix = (variable_prefix,)

    if lower_bound and upper_bound:
        diff = upper_bound - lower_bound
        if upper_bound <= 0:
            # invert to get the lowest possible constant and shift
            return -1 * get_binary_representation_polynomial(diff, 0, None, variable_prefix) + upper_bound
        if split_signs:
            positive_part = get_binary_representation_polynomial(upper_bound, 0, None, variable_prefix + ("+",))
            negative_part = -1 * get_binary_representation_polynomial(-lower_bound, 0, None, variable_prefix + ("-",))
            return positive_part + negative_part
        # shift bounds to use as few variables as possible
        return get_binary_representation_polynomial(diff, 0, None, variable_prefix) + lower_bound

    num_variables = num_variables or int(upper_bound).bit_length() - 1
    poly = PolyBinary({((*variable_prefix, i),): 2 ** i for i in range(num_variables)})
    if not upper_bound:
        return poly

    remaining = upper_bound - 2 ** num_variables + 1
    return poly + PolyBinary({((*variable_prefix, num_variables),): remaining})

def get_unary_representation_polynomial(upper_bound, lower_bound=0, variable_prefix=(X,)):
    """
    get the linear polynomial representing an integer variable bounded between the lower bound and the upper bound
    using the unary representation with binary variables,
    the polynomial will be empty if upper_bound == lower_bound == 0

    :param (Integer) upper_bound: value of the upper bound, can be 0 but not negative unless the lower bound is also set
    :param (Integer) lower_bound: value of the lower bound
    :param (tuple or str) variable_prefix: prefix for the binary variables that are created
    :return: the linear polynomial representing the integer variable
    """
    if upper_bound and upper_bound < lower_bound:
        raise ValueError(ERROR_UPPER)

    if not isinstance(variable_prefix, tuple):
        variable_prefix = (variable_prefix,)

    if upper_bound < 0:
        # without the mirroring to the negative numbers the constant would be much larger
        return -1 * get_unary_representation_polynomial(-1 * lower_bound, -1 * upper_bound, variable_prefix)
    poly = PolyBinary({((*variable_prefix, i),): 1 for i in range(upper_bound - lower_bound)})
    if lower_bound:
        poly += lower_bound
    return poly



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for IO with txt files """

import os


ERROR_ANY = "The file does not contain any object of type '{}'"
ERROR_FEW = "The file does not contain so much objects of type '{}'"


def save(obj, filename, mode="w"):
    """
    save object in txt file (by default only one object per file)

    :param obj: the object to be saved
    :param (str) filename: the name of txt file
    :param (str) mode: append to file ('a') or overwrite file ('w')
    """
    with open(filename, mode=mode, encoding="utf-8") as txt_file:
        if os.stat(filename).st_size > 0:
            txt_file.write("\n")
        # file should start with class name
        txt_file.write(type(obj).__name__ + "\n")
        # get the corresponding function for writing and call it
        obj.write_txt(txt_file)
        txt_file.write("\n")


def load(cls, filename, index=0):
    """
    load object of the given class from txt file

    :param cls: the type of the object to be loaded
    :param (str) filename: the name of txt file
    :param (int) index: if several objects of the same type are stored,
                        take the one with appears at index, by default the first
    """
    with open(filename, "r", encoding="utf-8") as txt_file:
        loaded_lines = txt_file.readlines()

    # search for correct place in the lines
    obj_lines = _get_lines(cls.__name__, loaded_lines, index)

    # get the corresponding read method and call it
    loaded_data = cls.read_txt(obj_lines)
    loaded_object = cls(**loaded_data)
    return loaded_object

def _get_lines(cls_name, lines, index=0):
    # remove artifacts
    lines = [line.replace("\n", "") for line in lines]

    start = -1
    end = None
    counter = -1
    for line_number, line in enumerate(lines):
        if line == cls_name:
            counter += 1
            if counter == index:
                # skip line with object name
                start = line_number + 1
        elif counter == index and line == "":
            end = line_number
            break

    if counter < 0:
        raise ValueError(ERROR_ANY.format(cls_name))
    if start < 0:
        raise ValueError(ERROR_FEW.format(cls_name))

    return lines[start:end]

def exists(cls, filename):
    """
    check if an object of the given class already exists in the txt file

    :param cls: the type of object to be checked
    :param (str) filename: the name of txt file
    :return: True if there is already an object of the given type in the file
    """
    if os.path.exists(filename):
        with open(filename, "r", encoding="utf-8") as txt_file:
            lines = txt_file.readlines()

        try:
            _get_lines(cls.__name__, lines)
            return True
        except ValueError:
            pass
    return False



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the Embedding """

import os
import pytest

from quark import Embedding, HardwareAdjacency
from quark.io import hdf5


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "embedding.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE

VAR_NODES_MAP = {0: [50, 51, 52], 1: [53, 54, 55, 56], 2: [70]}
VAR_EDGES_MAP = {0: [(50, 51), (51, 52)], 1: [(53, 54), (54, 55), (55, 56)], 2: []}
COUPLING_EDGES_MAP = {(0, 1): [(50, 53)], (1, 2): []}
EMBEDDING = Embedding(VAR_NODES_MAP, COUPLING_EDGES_MAP, VAR_EDGES_MAP, name="test_embedding")
EMBEDDING_INDEX = Embedding(VAR_NODES_MAP, name="test_embedding_index", index=1)

VAR_NODES_MAP_CHIMERA = {0: [4, 0, 24, 48],
                         1: [5, 1, 25, 49],
                         2: [6, 2, 26, 50],
                         3: [7, 3, 27, 51],
                         4: [28, 36, 32, 56],
                         5: [29, 37, 33, 57],
                         6: [30, 38, 34, 58],
                         7: [31, 39, 35, 59],
                         8: [68, 64, 60, 52],
                         9: [69, 65, 61, 53],
                         10: [70, 66, 62, 54],
                         11: [71, 67, 63, 55]}


def test_load_hdf5():
    """ test loading from h5 files """
    loaded = hdf5.load(Embedding, FILENAME_FIX, name=EMBEDDING.name)
    assert loaded == EMBEDDING

def test_io_hdf5():
    """ test io round trip """
    hdf5.save(EMBEDDING, FILENAME_TMP)
    loaded = hdf5.load(Embedding, FILENAME_TMP, name=EMBEDDING.name)
    assert loaded == EMBEDDING

    hdf5.save(EMBEDDING_INDEX, FILENAME_TMP)
    loaded = hdf5.load(Embedding, FILENAME_TMP, name=EMBEDDING_INDEX.name, index=EMBEDDING_INDEX.index)
    assert loaded == EMBEDDING_INDEX

    os.remove(FILENAME_TMP)

def test_complete_graph_embedding():
    """ test complete graph check """
    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        chimera_hwa = hdf5.load(HardwareAdjacency,
                                TEST_DIR + "testdata/hwa.h5",
                                full_group_name="ChimeraHWA/full_chimera_3_3_4")
    embedding = Embedding.get_from_hwa(VAR_NODES_MAP_CHIMERA, chimera_hwa)
    assert embedding.are_all_couplings_mapped('complete')



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ConstraintBinary """

from numbers import Real
from warnings import warn, catch_warnings, warn_explicit

from .poly_binary import PolyBinary, Polynomial, get_binary_representation_polynomial, get_all_reduction_penalty_terms
from .utils.variables import to_string


CONSTRAINT = "constraint"
SLACK      = "slack"
REDUCTION  = "reduction"
X          = "x"
ORIGINAL   = "original"

ERROR_CONSTANT      = "Constant must be a number"
ERROR_BOTH          = "At least one bound needs to be given"
ERROR_LOWER         = "Lower bound must be less than or equal to upper bound"
ERROR_UNFULFILLABLE = "Unfulfillable constraint '{}'"
ERROR_INT           = "Inequality constraint must have integer coefficients and boundaries"
ERROR_POSITIVE      = "Positive coefficients do not suffice to get above the lower bound"
ERROR_NEGATIVE      = "Negative coefficients do not suffice to get below the upper bound"
ERROR_BINARY        = "At the moment can only handle PolyBinary"
ERROR_LINEAR        = "Method only makes sense for linear polynomials"

WARNING_EQUAL_VARS = "The two variables '{}' and '{}' are set to be equal, consider replacing one with the other"
WARNING_NEG_VARS   = "The two variables '{}' and '{}' are set to be negations of each other, " \
                     "consider replacing one with '1 - other'"
WARNING_USELESS    = "Useless constraint '{}'"
WARNING_FLAT       = "Flat variables are replaced by tuple variables"
WARNING_ADVANCE    = "Variables could be set in advance: {}"
WARNING_SHARPEN    = "{} bound could be sharpened to {}"
WARNING_FORCE      = "Force is only necessary to get a fully linear problem"


class ConstraintBinary:
    """
    A constraint restricts the values of a function, here a polynomial.
    We define an inequality constraint with lower_bound <= polynomial <= upper_bound,
    where we have an equality constraint if lower_bound == upper_bound.
    """

    def __init__(self, polynomial, lower_bound=None, upper_bound=None, _check=True):
        """
        initialize ConstraintBinary object,
        if only one bound is given, the other is calculated by worst case estimation,
        note that this can result in hugh feasibility intervals,
        which lead to a large number of slack variables in the QUBO formulation

        :param (PolyBinary) polynomial: binary polynomial which is bounded
        :param (Real or None) lower_bound: value of the lower bound
        :param (Real or None) upper_bound: value of the upper bound
        :param (bool) _check: whether all the consistencies are checked, switch off only with care!
         """
        self._check_input(polynomial, lower_bound, upper_bound)
        self.polynomial = polynomial
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        if self.lower_bound is None:
            # worst case estimation of lower bound
            self.lower_bound = sum(polynomial.negative.values()) + polynomial.offset
        if upper_bound is None:
            # worst case estimation of upper bound
            self.upper_bound = sum(polynomial.positive.values()) + polynomial.offset

        if _check:
            self.check_consistency()

    @staticmethod
    def _check_input(polynomial, lower_bound, upper_bound):
        """ check whether the input is correct """
        if lower_bound is None and upper_bound is None:
            raise ValueError(ERROR_BOTH)
        if not isinstance(polynomial, PolyBinary):
            raise ValueError(ERROR_BINARY)

    def check_consistency(self):
        """ check if the data is consistent and defines a meaningful constraint """
        if self.lower_bound > self.upper_bound:
            raise ValueError(ERROR_LOWER)
        if self.polynomial.is_constant():
            if not self.lower_bound <= self.polynomial.offset <= self.upper_bound:
                # the polynomial only consisting of a constant is not in between the bounds
                raise ValueError(ERROR_UNFULFILLABLE.format(self))
            # if the constant is between the bounds, it is simply a useless constraint
            warn(WARNING_USELESS.format(self))
            return

        # check edge cases by sums of all positive and all negative coefficients
        # lb <= p+(x) + p-(x) + c <= ub
        if sum(self.polynomial.positive.values()) + self.polynomial.offset < self.lower_bound:
            # lb > p+(1) + c (can only happen if offset c is negative)
            raise ValueError(ERROR_POSITIVE)
        if sum(self.polynomial.negative.values()) + self.polynomial.offset > self.upper_bound:
            # p-(1) + c > ub (can only if offset c is positive, thus no contradiction with above)
            raise ValueError(ERROR_NEGATIVE)

        # other preprocessable cases
        self.get_preprocessable_variables(_raise_warning=True)

        # the reversed edge cases can be used to check the strengths of the bounds
        if (sum(self.polynomial.positive.values()) + self.polynomial.offset <= self.upper_bound) \
                and (sum(self.polynomial.negative.values()) + self.polynomial.offset >= self.lower_bound):
            # lb <= p-(1) + c <= p+(x) + p-(x) + c <= p+(1) + c <= ub
            warn(WARNING_USELESS.format(self))
        elif sum(self.polynomial.positive.values()) + self.polynomial.offset < self.upper_bound:
            warn(WARNING_SHARPEN.format("Upper", sum(self.polynomial.positive.values()) + self.polynomial.offset))
        elif sum(self.polynomial.negative.values()) + self.polynomial.offset > self.lower_bound:
            warn(WARNING_SHARPEN.format("Lower", sum(self.polynomial.negative.values()) + self.polynomial.offset))

    def __add__(self, constant):
        """
        shift the constraint by adding the constant to the polynomial, the lower_bound and the upper_bound

        :param (Real) constant: constant value to be added
        :return: the shifted constraint
        """
        if not isinstance(constant, Real):
            raise ValueError(ERROR_CONSTANT)
        return ConstraintBinary(self.polynomial + constant, self.lower_bound + constant, self.upper_bound + constant,
                                _check=False)

    def __sub__(self, constant):
        """
        shift the constraint by subtracting the constant from the polynomial, the lower_bound and the upper_bound

        :param (Real) constant: constant value to be subtracted
        :return: the shifted constraint
        """
        return self.__add__(constant * -1)

    def __mul__(self, constant):
        """
        scale the constraint by multiplying the constant to the polynomial, the lower_bound and the upper_bound

        :param (Real) constant: constant value to be multiplied
        :return: the scaled constraint
        """
        if not isinstance(constant, Real) or constant == 0:
            raise ValueError(ERROR_CONSTANT)
        lower_bound, upper_bound = self.lower_bound * constant, self.upper_bound * constant
        if constant < 0:
            lower_bound, upper_bound = upper_bound, lower_bound
        return ConstraintBinary(self.polynomial * constant, lower_bound, upper_bound, _check=False)

    def __eq__(self, other):
        """
        check equality of constraints, in the polynomial monomials with zero coefficient are not considered,
        as well as possible shifting and scaling of the constraints
        """
        if not isinstance(other, ConstraintBinary):
            return False
        if self.polynomial.var_type != other.polynomial.var_type:
            return False

        # direct comparison check
        normalized_self = self.normalize()
        normalized_other = other.normalize()
        if normalized_self.polynomial == normalized_other.polynomial \
            and normalized_self.upper_bound == normalized_other.upper_bound:
            return True
        # shifting/scaling check
        if normalized_self.polynomial:
            var_tuple, coeff_self = next(iter(normalized_self.polynomial.items()))
            if var_tuple not in normalized_other.polynomial:
                return False
            poly_scale = coeff_self / normalized_other.polynomial[var_tuple]
            return normalized_self.polynomial == normalized_other.polynomial * poly_scale \
                   and normalized_self.upper_bound == normalized_other.upper_bound * poly_scale
        return False

    def __repr__(self):
        """ get nicely formatted representation of the constraint """
        return self.__class__.__name__ + f"({repr(self.polynomial)}, {self.lower_bound}, {self.upper_bound})"

    def __str__(self, revert_eq=False):
        """ get human-readable representation of the constraint """
        if self.is_equality_constraint() and not revert_eq:
            return f"{self.polynomial} == {self.upper_bound:n}"
        if self.is_equality_constraint():
            return f"{self.upper_bound:n} == {self.polynomial}"
        return f"{self.lower_bound:n} <= {self.polynomial} <= {self.upper_bound:n}"

    def copy(self):
        """ get a copy of the constraint """
        return ConstraintBinary(self.polynomial.copy(), self.lower_bound, self.upper_bound, _check=False)

    def is_equality_constraint(self):
        """ check if the constraint is an equality constraint """
        return self.lower_bound == self.upper_bound

    def is_integer(self, precision=1e-9):
        """
        check if the lower bound, the upper bound and the coefficients of the polynomial are all integer

        :param (Real) precision: precision value until which values are considered to be integer
        :return: True if all values are integer
        """
        return self.polynomial.has_int_coefficients(precision) \
               and abs(self.lower_bound - int(self.lower_bound)) < precision \
               and abs(self.upper_bound - int(self.upper_bound)) < precision

    def is_normal(self):
        """ check if the constraint is in normalized format """
        return self.lower_bound == 0 and _is_first_monomial_non_negative(self.polynomial)

    def round(self, decimal_cap=None):
        """
        round the coefficients of the polynomial and the bounds of the constraint according to the decimal cap

        :param (int or None) decimal_cap: the number of digits of the resulting coefficients,
                                          if None, they will be ints
        :return: the new constraint with rounded bounds and polynomial
        """
        rounded_lower_bound = round(self.lower_bound, decimal_cap)
        rounded_upper_bound = round(self.upper_bound, decimal_cap)
        return self.__class__(self.polynomial.round(decimal_cap), rounded_lower_bound, rounded_upper_bound)

    def normalize(self):
        """
        get the constraint with the normalized format where lower_bound=0
        by subtracting the lower bound
        """
        if self.is_normal() and not any(coeff == 0 for coeff in self.polynomial.values()):
            return self
        constraint = self if _is_first_monomial_non_negative(self.polynomial) else (self * -1)
        constraint -= constraint.lower_bound
        constraint.polynomial = constraint.polynomial.remove_zero_coefficients()
        return constraint

    def replace_variables(self, replacement, _check=True):
        """
        replace the variables in the polynomial with other variables

        :param (list or dict or function) replacement: mapping to the new variables, if replacement is a dictionary, it
                                                       should be of format {old_var : new_var}, can only be a list if
                                                       the polynomial is flat
        :param (bool) _check: whether all the consistencies are checked, switch off only with care!
        :return: the new ConstraintBinary with replaced variables
        """
        return ConstraintBinary(self.polynomial.replace_variables(replacement), self.lower_bound, self.upper_bound,
                                _check=_check)

    def preprocess(self):
        """ replace preprocessable variables, if existing, in the constraint polynomial """
        preprocessable = self.get_preprocessable_variables()
        new_poly = self.polynomial.evaluate(preprocessable)
        new_poly = PolyBinary(new_poly).remove_zero_coefficients()

        if new_poly.is_constant() and self.lower_bound <= new_poly.offset <= self.upper_bound:
            # fully processed constraint -> remaining part would be a useless constraint
            return None, preprocessable
        return ConstraintBinary(new_poly, self.lower_bound, self.upper_bound), preprocessable

    def get_preprocessable_variables(self, _raise_warning=False):
        """
        get the variables in the polynomial that can be preprocessed
        due to a specific structure of the constraint
        """
        normalized_constrained = self.normalize()

        if self.is_equality_constraint() and len(self.polynomial.variables) == 2:
            preprocessable = _get_preprocessable_variables_eq_or_neg(normalized_constrained.polynomial, _raise_warning)
            if preprocessable:
                return preprocessable

        preprocessable0, preprocessable1 = _get_preprocessable_variables_bounds(normalized_constrained.polynomial,
                                                                                normalized_constrained.upper_bound)
        if not set(preprocessable0).isdisjoint(preprocessable1):
            # if one variable should be set to two different values, there is a contradiction in the constraint
            raise ValueError(ERROR_UNFULFILLABLE.format(self))

        preprocessable0.update(preprocessable1)
        if _raise_warning and preprocessable0:
            warn(WARNING_ADVANCE.format(preprocessable0))
        return preprocessable0

    def to_equality_constraint(self, slack_variable_prefix=(SLACK,)):
        """
        transform the constraint into an equality constraint by automatically introducing slack variables if necessary

        :param (tuple or str) slack_variable_prefix: prefix for the additional slack variables that are created
        :return: the equality constraint
        """
        normalized_constraint = self.normalize()
        if self.is_equality_constraint():
            return normalized_constraint
        if not isinstance(slack_variable_prefix, tuple):
            slack_variable_prefix = (slack_variable_prefix,)

        slack_polynomial = get_binary_representation_polynomial(normalized_constraint.upper_bound,
                                                                variable_prefix=slack_variable_prefix)
        return ConstraintBinary(normalized_constraint.polynomial - slack_polynomial, 0, 0)

    def get_reductions(self, max_degree=1, use=None, force=False, reduction_strategy=None):
        """
        get the reduced constraint and the corresponding reductions,
        if the constraint encodes a reduction itself, we also retrieve this if forced

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (list[tuples] or None) use: the reductions which shall be used,
                                           as a list with tuples in the format (var1, var2, reduction_var)
        :param (bool) force: if True also reduce quadratic polynomials to linear that could remain for the QUBO problem
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :return the reduced constraint, a list of variable triples representing the reductions
        """
        if (not max_degree or max_degree > 1) and force:
            warn(WARNING_FORCE)
            force = False

        # TODO: reduce number of calls of _match_special_constraint_get_penalty_poly
        if max_degree and not force and self.provides_penalty_term_directly(None):
            # if the constraint polynomial already provides a valid penalty term of any degree,
            # we only reduce its degree below 2 if forced
            max_degree = max(max_degree, 2)

        # check whether the constraint encodes a reduction itself
        reduction = retrieve_reduction(self)
        if reduction:
            if force or (max_degree and max_degree == 1 and self.polynomial.degree >= 2):
                # if we are forced or if we have the quadratic reduction constraint r == x * y and want linear
                return None, [reduction]
            # otherwise we keep the linear or quadratic constraints
            return self, []

        # use already given reductions
        reduced_poly, _ = self.polynomial.reduce(None, use, reduction_strategy=reduction_strategy)
        if max_degree and reduced_poly == self.polynomial and self.polynomial.degree <= max_degree:
            # nothing has changed and has to be changed
            return self, []

        # when this point is reached, simply reduce to given max degree
        reduced_poly, reductions = reduced_poly.reduce(max_degree, reduction_strategy=reduction_strategy)
        with catch_warnings(record=True) as warnings:
            reduced_constraint = ConstraintBinary(reduced_poly, self.lower_bound, self.upper_bound)
        for warning in warnings:
            if issubclass(warning.category, UserWarning) and warning.message.args[0].startswith("Useless constraint"):
                # in case we have created a useless constraint, we do not return it
                reduced_constraint = None
            else:
                warn_explicit(message=warning.message, category=warning.category, filename=warning.filename,
                              lineno=warning.lineno, source=warning.source)
        return reduced_constraint, reductions

    def reduce(self, max_degree=1, use=None, force=False, reduction_strategy=None, reduced_constraint_name=ORIGINAL):
        """
        get a new constraint with just a polynomial with at most the given degree by substituting products of variables

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (list[tuples] or None) use: the reductions which shall be used,
                                           as a list with tuples in the format (var1, var2, reduction_var)
        :param (bool) force: if True also reduce quadratic polynomials to linear that could remain for the QUBO problem
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :param (str) reduced_constraint_name: the name of the created reduced constraint
        :return a dictionary with the reduced constraint and new constraints enforcing the reductions
        """
        reduced_constraint, reductions = self.get_reductions(max_degree, use, force, reduction_strategy)
        if not max_degree:
            assert not reductions
        constraints = {reduced_constraint_name: reduced_constraint} if reduced_constraint else {}
        for reduction in reductions:
            constraints.update(get_reduction_constraints(*reduction, max_degree=max_degree))
        return constraints

    def provides_penalty_term_directly(self, max_degree=2):
        """
        check whether the polynomial can be used itself as a penalty term

        :param (int or None) max_degree: the maximum degree to be allowed
        :return: whether the polynomial of the constraints directly provides a penalty polynomial
        """
        if not self.is_equality_constraint() or (max_degree and self.polynomial.degree > max_degree):
            return False

        normalized_constraint = self.normalize()
        if normalized_constraint.polynomial.is_non_negative():
            # if the polynomial is >=0 for all assignments and shall evaluate to 0,
            # it can be used itself as a penalty term
            assert normalized_constraint.polynomial.offset == 0  # invalid constructions should be caught by init
            return True

        return matches_special_penalty_constraint(normalized_constraint)

    def get_penalty_terms(self, name=CONSTRAINT, int_precision=1e-9, reduction_strategy=None,
                          check_special_constraints=True):
        """
        get a quadratic polynomial term which represents the constraint,
        that is, it evaluates to 0 if the constraint is fulfilled and to a positive value otherwise,
        (thus penalizes invalid variable assignments)

        if necessary, slack variables are added and/or a quadratic polynomial is reduced to a linear one,
        the latter causes additional penalty term(s), such that we get, e.g.,
            {name: penalty term directly derived from constraint,
            'reduction_x_1_y_1': penalty term for reduction of product x_1*y_1,
            ...}

        :param (str) name: name of the penalty term directly derived from constraint
        :param (Real) int_precision: precision value for checking if the polynomial is integer, which is needed in case
                                     of an inequality constraint, because those can only be handled if they are integer
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :param (bool) check_special_constraints: if true, checks whether there are special constraints
                                                 that can be handled more efficiently than by standard transformation
        :return: the dictionary with names to terms
        """
        # TODO: add option to get penalty terms without prior reduction -> set max degree
        if self.polynomial.is_constant():
            # in this case the constraint is not actually a constraint
            # invalid constructions should be caught by initialization
            assert self.lower_bound <= self.polynomial.offset <= self.upper_bound
            return {}
        if not (self.is_integer(int_precision) or self.is_equality_constraint()):
            # we can only handle integer inequality constraints or arbitrary equality constraints
            raise ValueError(ERROR_INT)

        if check_special_constraints:
            penalty_poly = get_special_constraint_penalty_poly(self)
            if penalty_poly and penalty_poly.is_quadratic():
                # for some special cardinality constraints, the penalty terms are not quadratic
                # for now, we have decided that those will be handled as normal linear constraints
                # where slack variables are introduced rather than reductions applied
                return {name: penalty_poly}

        reduced_constraint, reductions = self.get_reductions(reduction_strategy=reduction_strategy)
        penalty_terms = get_all_reduction_penalty_terms(reductions)

        if reduced_constraint.provides_penalty_term_directly():
            penalty_poly = reduced_constraint.polynomial - reduced_constraint.polynomial.offset
        else:
            penalty_poly = get_standard_penalty_polynomial(reduced_constraint, (name + '_' + SLACK,))

        penalty_terms.update({name: penalty_poly})
        return penalty_terms

    def check_validity(self, var_assignment):
        """
        check if a given variable assignment fulfills the constraint

        :param (dict or list) var_assignment: mapping from variables to values
        :return: True if the constraint is fulfilled
        """
        poly_value = self.polynomial.evaluate(var_assignment)
        return self.lower_bound <= poly_value <= self.upper_bound


# helper methods

## normalization

def _is_first_monomial_non_negative(polynomial):
    """
    check if the polynomial has a non-negative first monomial (not being the offset)

    :param (PolyBinary) polynomial:
    :return: True if the first monomial has a non-negative coefficient
    """
    if polynomial.is_constant():
        return True
    polynomial = polynomial.remove_zero_coefficients()
    polynomial.sort_entries()
    first_monomial = list(polynomial.keys())[0] or list(polynomial.keys())[1]  # if the first is the offset take second
    return polynomial[first_monomial] >= 0

## preprocessing

POLY_EQUATION = PolyBinary({(0,): 1, (1,): -1})
POLY_NEGATION = PolyBinary({(0,): 1, (1,): 1, (): -1})

def _get_preprocessable_variables_eq_or_neg(polynomial, _raise_warning=False):
    """ get the preprocessable variables for polynomials enforcing an equality or a negation of variables """
    compact_poly = polynomial.compact()
    if compact_poly == POLY_EQUATION:  # case x == y
        if _raise_warning:
            warn(WARNING_EQUAL_VARS.format(polynomial.variables[0], polynomial.variables[1]))
        return {polynomial.variables[1]: Polynomial({(polynomial.variables[0],): 1})}

    if compact_poly == POLY_NEGATION:  # case x == 1 - y
        if _raise_warning:
            warn(WARNING_NEG_VARS.format(polynomial.variables[0], polynomial.variables[1]))
        return {polynomial.variables[1]: Polynomial({(polynomial.variables[0],): -1, (): 1})}

    return None

def _get_preprocessable_variables_bounds(polynomial, upper_bound):
    # want 0 <= p+(x) + p-(x) + c <= ub, have p-(1) <= p+(x) + p-(x) <= p+(1) for all x
    # invalid constructions should be caught by initialization
    assert sum(polynomial.positive.values()) + polynomial.offset >= 0
    if sum(polynomial.positive.values()) + polynomial.offset == 0:
        # p+(1) + c == 0 -- this means we are in general below the lower bound and only hit it in the extreme case
        # -> all summands in p+ need to be set to 1 and all summands in p- to 0
        # -> all variables of p+ need to be set to 1 and all variables of p-.linear need to be set to 0
        to_1 = {var: 1 for var in polynomial.positive.variables}
        to_0 = {var: 0 for var in polynomial.negative.linear.variables}
        return to_0, to_1

    assert sum(polynomial.negative.values()) + polynomial.offset <= upper_bound
    if sum(polynomial.negative.values()) + polynomial.offset == upper_bound:
        # p-(1) + c == ub -- this means we are in general above the upper bound and only hit it in the extreme case
        # -> all summands in p- need to be set to 1 and all summands in p+ to 0
        # -> all variables of p- need to be set to 1 and all variables of p+.linear need to be set to 0
        to_1 = {var: 1 for var in polynomial.negative.variables}
        to_0 = {var: 0 for var in polynomial.positive.linear.variables}
        return to_0, to_1

    return {}, {}

## standard linear constraint reformulation

def get_standard_penalty_polynomial(constraint, slack_variable_prefix=(SLACK,)):
    """
    get the squared polynomial penalty term representing the constraint,
    automatically introduces slack variables in case it is an inequality constraint

    this is done by transforming the inequality to an equality with 'polynomial == s' for '0 <= s <= upper_bound'
    where s is an integer slack variable, which gets replaced by a binary representation,
    then the penalty term is formed with '(polynomial - s)**2',
    if the upper bound is 0, we have no slack variables

    note that this only produces proper QUBO penalty terms when the polynomial of the constraint was linear

    :param (ConstrainedBinary) constraint: constraint which shall be transformed
    :param (tuple or str) slack_variable_prefix: prefix for the additional slack variables that are created
    :return: the polynomial representing the constraint
    """
    eq_constraint = constraint.to_equality_constraint(slack_variable_prefix)
    return eq_constraint.polynomial ** 2

## special constraints

### reduction constraints

#### standard linear reduction constraints
REDUCTION_CONSTRAINT_POLY_LIN2 = PolyBinary({(0,): 1, (1,): -1, (): 1})  # 2 variables
REDUCTION_CONSTRAINT_POLY_LIN3 = PolyBinary({(0,): 1, (1,): -1, (2,): -1, (): 1})  # 3 variables

REDUCTION_CONSTRAINT_LIN2 = ConstraintBinary(REDUCTION_CONSTRAINT_POLY_LIN2, 0, 1)  # x0 <=  x1
                                                                                # <=> (0 <=) x0 - x1 + 1 <= 1
REDUCTION_CONSTRAINT_LIN3 = ConstraintBinary(REDUCTION_CONSTRAINT_POLY_LIN3, 0, 1)  # x0 >= x1 + x2 - 1
                                                                             # <=>     0 <=   x0 - x1 - x2 + 1 (<= 2/1)
                                                                             # <=> (-1/0 <=) -x0 + x1 + x2      <= 1

#### original quadratic reduction constraint
REDUCTION_CONSTRAINT_POLY_QUAD  = PolyBinary({(0,): 1, (1, 2): -1})
REDUCTION_CONSTRAINT_QUAD = ConstraintBinary(REDUCTION_CONSTRAINT_POLY_QUAD, 0, 0)  # x0 == x1 * x2
                                                                                 # <=> 0 == x0 - x1 * x2

#### corresponding penalty polynomials, summing up to standard full reduction penalty polynomial
REDUCTION_PENALTY_POLY_LIN2 = PolyBinary({(0,): 1, (0, 1): -1})                        #    x0 -  x0 x1
                                              # analogously when second variable is 2  #    x0          -  x0 x2
REDUCTION_PENALTY_POLY_LIN3 = PolyBinary({(0,): 1, (0, 1): -1, (0, 2): -1, (1, 2): 1}) #    x0 -  x0 x1 -  x0 x2 + x1 x2
                                                                                       # ===============================
REDUCTION_PENALTY_POLY_FULL = PolyBinary({(0,): 3, (0, 1): -2, (0, 2): -2, (1, 2): 1}) # +3 x0 -2 x0 x1 -2 x0 x2 + x1 x2

#### constraints with penalty polynomials
REDUCTION_CONSTRAINT_PENALTY_L2 = ConstraintBinary(REDUCTION_PENALTY_POLY_LIN2, 0, 0)
REDUCTION_CONSTRAINT_PENALTY_L3 = ConstraintBinary(REDUCTION_PENALTY_POLY_LIN3, 0, 0)
REDUCTION_CONSTRAINT_PENALTY_FULL = ConstraintBinary(REDUCTION_PENALTY_POLY_FULL, 0, 0)

CONSTRAINT_2_PENALTY_POLY = [(REDUCTION_CONSTRAINT_LIN2,         REDUCTION_PENALTY_POLY_LIN2, False),
                             (REDUCTION_CONSTRAINT_PENALTY_L2,   REDUCTION_PENALTY_POLY_LIN2, True),
                             (REDUCTION_CONSTRAINT_LIN3,         REDUCTION_PENALTY_POLY_LIN3, False),
                             (REDUCTION_CONSTRAINT_QUAD,         REDUCTION_PENALTY_POLY_FULL, False),
                             (REDUCTION_CONSTRAINT_PENALTY_L3,   REDUCTION_PENALTY_POLY_LIN3, True),
                             (REDUCTION_CONSTRAINT_PENALTY_FULL, REDUCTION_PENALTY_POLY_FULL, True)]

def get_reduction_constraints(reduction_var, var1, var2=None, max_degree=1):
    """
    get the constraint(s) that enforce a variable to be equal to the product of two variables
    (the variables need to be of the same type, thus either tuple, int or str),
    if max_degree is 1, we add the 3 linearization constraints,
    if is greater than 1, then we simply add the constraint var1 * var2 = reduction_var,

    Remark: In the traditional linearization strategy, the lower bound of the third constraint is unset
            (warning.e., the lowest possible value of the polynomial would be -1,
            which our constraint would also default to).
            Here we set the lower bound to 0, which additionally excludes another variable constellation.
            This case is actually already excluded by the other constraints.
            But this setting allows later to handle the below constraints individually and
            nevertheless get the correct quadratic penalty term for the full reduction
            by adding the individual penalty terms.

    :param (tuple or int or str) reduction_var: new binary variable that substitutes var1 * var2
    :param (tuple or int or str) var1: first binary variable
    :param (tuple or int or str) var2: second binary variable
    :param (int) max_degree: the maximum degree of the constraint polynomials, by default 1
    :return: the constraint(s) that enforce var1 * var2 = reduction_var
    """
    if var2 is None:
        return {}  # TODO: return the single corresponding reduction variable here?
    name = to_string(reduction_var)
    if max_degree > 1:
        poly = REDUCTION_CONSTRAINT_POLY_QUAD.replace_variables([reduction_var, var1, var2])
        return {name: ConstraintBinary(poly, 0, 0)}

    name += "_{}"
    polys = [REDUCTION_CONSTRAINT_POLY_LIN2.replace_variables([reduction_var, var1]),
             REDUCTION_CONSTRAINT_POLY_LIN2.replace_variables([reduction_var, var2]),
             REDUCTION_CONSTRAINT_POLY_LIN3.replace_variables([reduction_var, var1, var2])]
    return {name.format(i): ConstraintBinary(poly, 0, 1) for i, poly in enumerate(polys)}

def _match_reduction_constraint_get_penalty_poly(constraint, only_penalty_constraints=False, return_variables=False):
    """ get the penalty polynomial (or the correspondingly permuted variables) corresponding to the constraint """
    normalized_constraint = constraint.normalize()
    variables = tuple(normalized_constraint.polynomial.variables)
    compact = ConstraintBinary(normalized_constraint.polynomial.compact(), 0, normalized_constraint.upper_bound,
                               _check=False)

    for permutation in [[0, 1, 2], [1, 0, 2], [2, 1, 0]]:  # 0 -> 1 -> 1, 1 -> 0 -> 2, 2 -> 2 -> 0
        constraint_permuted = compact.replace_variables(permutation, _check=False)
        variables_permuted = tuple(variables[i] for i in permutation if i < len(variables))
        for reduction_constraint, reduction_penalty_poly, is_penalty in CONSTRAINT_2_PENALTY_POLY:
            if (is_penalty or not only_penalty_constraints) and constraint_permuted == reduction_constraint:
                if return_variables:
                    return variables_permuted
                return reduction_penalty_poly.replace_variables(variables_permuted)

    return None

def retrieve_reduction(constraint):
    """
    check if the constraint corresponds to a reduction and if yes recover the reduction

    :param constraint: the constraint to recover the corresponding reduction from
    :return: the reduction variable triple
    """
    variables = _match_reduction_constraint_get_penalty_poly(constraint, return_variables=True)
    if variables:
        # we just sort the two variables that are reduced
        return (variables[0],) + tuple(sorted(variables[1:]))
    return None

### check for all special constraints (more are to be implemented)

def get_special_constraint_penalty_poly(constraint):
    """
    get the corresponding penalty polynomial if the given constraint fits to one of then special ones

    :param constraint: the constraint to be checked
    :return: the corresponding penalty polynomial
    """
    penalty_poly = _match_reduction_constraint_get_penalty_poly(constraint)
    if penalty_poly:
        return penalty_poly
    return None

def matches_special_penalty_constraint(constraint):
    """
    check whether the constraint matches a penalty polynomial of a special constraint

    :param constraint: the constraint to be checked
    """
    if _match_reduction_constraint_get_penalty_poly(constraint, only_penalty_constraints=True):
        return True
    return False



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the Instance """

import os
from abc import ABC
import pytest

from quark.io import hdf5, Instance
from quark.testing import ExampleInstance


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "instance.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE


def test_consistency(instance):
    """ test consistency checks """
    with pytest.raises(ValueError, match="Instance is not consistent: No edges given"):
        ExampleInstance([], len(instance.colors))
    with pytest.raises(ValueError, match="Instance is not consistent: No colors given"):
        ExampleInstance(instance.edges, 0)
    with pytest.raises(ValueError, match="Instance is not consistent: Edges need to have two nodes"):
        ExampleInstance([(), (), ()], len(instance.colors))

def test_get_name(instance):
    """ test instance name """
    assert instance.get_name() == "ExampleInstance_density_0.7_colors_3"

def test_io_hdf5(instance):
    """ test hdf5 io round trip """
    hdf5.save(instance, FILENAME_TMP)
    assert hdf5.exists(ExampleInstance, FILENAME_TMP)
    loaded = hdf5.load(ExampleInstance, FILENAME_TMP)
    assert loaded.edges == instance.edges
    assert loaded.colors == instance.colors
    os.remove(FILENAME_TMP)

    instance.save_hdf5(FILENAME_TMP)
    assert ExampleInstance.exists_hdf5(FILENAME_TMP)
    loaded = ExampleInstance.load_hdf5(FILENAME_TMP)
    assert loaded.edges == instance.edges
    assert loaded.colors == instance.colors
    os.remove(FILENAME_TMP)

def test_load_hdf5(instance):
    """ test hdf5 io """
    assert hdf5.exists(ExampleInstance, FILENAME_FIX)
    loaded = hdf5.load(ExampleInstance, FILENAME_FIX)
    assert loaded.edges == instance.edges
    assert loaded.colors == instance.colors

    assert ExampleInstance.exists_hdf5(FILENAME_FIX)
    loaded = ExampleInstance.load_hdf5(FILENAME_FIX)
    assert loaded.edges == instance.edges
    assert loaded.colors == instance.colors

class ExampleInstanceWithoutIO(Instance, ABC):
    """ some test instance """
    def check_consistency(self):
        """ test if instance is consistent """
        return True

def test_not_implemented():
    """ test not implemented methods """
    with pytest.raises(NotImplementedError):
        Instance().check_consistency()
    with pytest.raises(NotImplementedError):
        ExampleInstanceWithoutIO().get_name()
    with pytest.raises(NotImplementedError):
        Instance.read_hdf5(FILENAME_FIX)
    with pytest.raises(NotImplementedError):
        ExampleInstanceWithoutIO().write_hdf5(FILENAME_FIX)

    # pylint: disable=protected-access
    with pytest.raises(NotImplementedError):
        ExampleInstanceWithoutIO().get_constrained_objective()
    with pytest.raises(NotImplementedError):
        ExampleInstanceWithoutIO().get_objective_terms()
    with pytest.raises(NotImplementedError):
        ExampleInstanceWithoutIO().get_objective()



# Copyright 2025 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing namedtuple variables """

import pytest

from quark import PolyBinary, PolyIsing
from quark.variables import var_binary_namedtuple, var_ising_namedtuple


X = var_binary_namedtuple("x", "node", "index")
Y = var_binary_namedtuple("y", "node")
S = var_ising_namedtuple("s", "item", "value", "weight")

def test_init():
    """ test initialization """
    assert X("a", 1) == ("x", "a", 1)
    assert X("a", 1).var == "x"
    assert X("a", 1).node == "a"  # pylint: disable=no-member
    assert X("a", 1).index == 1   # pylint: disable=comparison-with-callable  # strange pylint false positives

    assert Y("b") == ("y", "b")
    assert S("box", 10, 5) == ("s", "box", 10, 5)

def test_add():
    """ test addition """
    assert X("a", 1) + X("a", 2) + Y("a") == PolyBinary({(('x', 'a', 1),): 1, (('x', 'a', 2),): 1, (('y', 'a'),): 1})
    assert X("a", 1) + 3 == PolyBinary({(('x', 'a', 1),): 1, (): 3})
    assert S("box", 10, 5) + S("ball", 2, 15) == PolyIsing({(("s", "box", 10, 5),): 1, (("s", "ball", 2, 15),): 1})

    with pytest.raises(ValueError, match="Cannot add PolyIsing and PolyBinary"):
        _ = X("a", 1) + S("box", 10, 5)

def test_radd():
    """ test addition from the right """
    assert 3 + X("a", 1) == PolyBinary({(('x', 'a', 1),): 1, (): 3})
    assert -7 + S("box", 10, 5) == PolyIsing({(("s", "box", 10, 5),): 1, (): -7})

def test_sub():
    """ test subtraction """
    assert X("a", 1) - X("a", 2) - Y("a") == PolyBinary({(('x', 'a', 1),): 1, (('x', 'a', 2),): -1, (('y', 'a'),): -1})
    assert X("a", 1) - 3 == PolyBinary({(('x', 'a', 1),): 1, (): -3})
    assert S("box", 10, 5) - S("ball", 2, 15) == PolyIsing({(("s", "box", 10, 5),): 1, (("s", "ball", 2, 15),): -1})

    with pytest.raises(ValueError, match="Cannot add PolyBinary and PolyIsing"):
        _ = X("a", 1) - S("box", 10, 5)

def test_rsub():
    """ test subtraction from the right """
    assert 3 - X("a", 1) == PolyBinary({(('x', 'a', 1),): -1, (): 3})
    assert -7 - S("box", 10, 5) == PolyIsing({(("s", "box", 10, 5),): -1, (): -7})

def test_mul():
    """ test multiplication """
    assert X("a", 1) * X("a", 2) * Y("a") == PolyBinary({(('x', 'a', 1), ('x', 'a', 2), ('y', 'a')): 1})
    assert X("a", 1) * 3 == PolyBinary({(('x', 'a', 1),): 3})
    assert S("box", 10, 5) * S("ball", 2, 15) == PolyIsing({(("s", "box", 10, 5), ("s", "ball", 2, 15)): 1})

    with pytest.raises(ValueError, match="Cannot multiply PolyBinary and PolyIsing"):
        _ = X("a", 1) * S("box", 10, 5)

def test_rmul():
    """ test multiplication from the right """
    assert 3 * X("a", 1) == PolyBinary({(('x', 'a', 1),): 3})
    assert -7 * S("box", 10, 5) == PolyIsing({(("s", "box", 10, 5),): -7})

def test_truediv():
    """ test division """
    assert X("a", 1) / 3 == PolyBinary({(('x', 'a', 1),): 1 / 3})
    assert S("box", 10, 5) / (-7) == PolyIsing({(("s", "box", 10, 5),): -1 / 7})

def test_pow():
    """ test power """
    assert X("a", 1) ** 3 == PolyBinary({(('x', 'a', 1),): 1})
    assert S("box", 10, 5) ** 3 == PolyIsing({(("s", "box", 10, 5),): 1})



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the VariableMapping """

from quark import VariableMapping
from quark.utils.variables import replace_in_tuples, replace_strs_by_ints, are_consecutive
from . import add, hdf5_datasets


VARIABLES             = "variables"
COMPOSITE_VARIABLES   = "composite_variables"
STRING_VARIABLE_PARTS = "string_variable_parts"

DATASETS = ["keys", "values"]


add.save_load_exists(VariableMapping)

@add.self_method(VariableMapping)
def write_hdf5(variable_mapping, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (VariableMapping) variable_mapping: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    for ds_name in DATASETS:
        variables = getattr(variable_mapping, ds_name)()
        if not are_consecutive(variables):
            write_variables(variables, group, "_" + ds_name)

def write_variables(variables, group, suffix=""):
    """ extract the parts of the variables and write them into the group """
    int_variables, int_to_str = replace_strs_by_ints(variables)
    hdf5_datasets.write_dataset(group, COMPOSITE_VARIABLES + suffix, dataset=int_variables)
    if int_to_str:
        hdf5_datasets.write_dataset(group, STRING_VARIABLE_PARTS + suffix, dataset=int_to_str)

@add.static_method(VariableMapping)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    variables = [read_variables(group, "_" + ds_name) for ds_name in DATASETS]
    if not variables[0] and variables[1]:
        return {VARIABLES: zip(range(len(variables[1])), variables[1])}
    if not variables[1] and variables[0]:
        return {VARIABLES: zip(variables[0], range(len(variables[0])))}
    assert all(variables)
    return {VARIABLES: zip(*variables)}

def read_variables(group, suffix=""):
    """ retrieve the parts of the variables """
    composite_variables = hdf5_datasets.read_dataset(group, COMPOSITE_VARIABLES + suffix, check_existence=False)
    if composite_variables is not None:
        composite_variables = list(tuple(int(i) for i in v) for v in composite_variables)
        int_to_str = hdf5_datasets.read_dataset(group, STRING_VARIABLE_PARTS + suffix, check_existence=False)
        if int_to_str:
            composite_variables = replace_in_tuples(composite_variables, int_to_str)
    return composite_variables



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the HardwareAdjacency """

import os

from quark import HardwareAdjacency
from quark.io import hdf5


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "hwa.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE

HWA = HardwareAdjacency([(5, 6), (1, 4), (3, 5)], "test_hwa_fix")


def test_load_hdf5():
    """ test loading from h5 files """
    loaded = hdf5.load(HardwareAdjacency, FILENAME_FIX, name=HWA.name)
    assert loaded == HWA

def test_io_hdf5():
    """ test IO round trip """
    hdf5.save(HWA, FILENAME_TMP)
    loaded = hdf5.load(HardwareAdjacency, FILENAME_TMP, name=HWA.name)
    assert loaded == HWA
    os.remove(FILENAME_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for IO of datasets with HDF5 files """

import itertools
from numbers import Real
import warnings
import h5py
import numpy as np


KEYS = "keys"
VALS = "values"

ERROR_SCALAR    = "Given dataset is just a scalar, create attribute instead"
ERROR_TYPE      = "Storing of dataset with type '{}' is not supported"
ERROR_NON       = "Storing of dataset of non-homogeneous type is not supported"
ERROR_LIST      = "Storing of dataset with type list over '{}' is not supported"
ERROR_KEYS      = "Dictionary keys are not of homogeneous type"
ERROR_ITEMS     = "Key items are not of homogeneous type"
ERROR_NOT_FOUND = "Did not find dataset '{}' in group '{}' in HDF5 file '{}'"
ERROR_MISMATCH  = "Mismatched dictionary dataset in group '{}' with {} keys but {} values"


def write_dataset(group, name, obj=None, dataset=None):
    """
    write the dataset in the open group of an HDF5 file,
    either the dataset is taken from the provided object with the attribute name or needs to be given explicitly

    :param (h5py.Group) group: the group in the HDF5 file in which the dataset is stored
    :param (str) name: the name under which the dataset is stored
    :param obj: the object with an attribute called as name, whose value is taken as the dataset to be stored
    :param dataset: the explicit dataset which shall be stored if no object is given
    """
    # if the object is given it should have an attribute which is called name
    if obj and hasattr(obj, name):
        dataset = getattr(obj, name)

    if name in group:
        del group[name]

    if not isinstance(dataset, dict):
        _write_flat_dataset(group, name, dataset)
    else:
        _write_dict_dataset(group, name, dataset)


def write_datasets(group, obj=None, *names, **names_to_datasets):  # pylint: disable=keyword-arg-before-vararg
    """
    write several datasets in the open group of an HDF5 file,
    either provide the names of the attributes of the given object yielding the datasets or
    give the explicit dictionary of names to datasets as keyword arguments

    :param (h5py.Group) group: the HDF5 group to store the datasets in
    :param obj: the object whose attributes to take as datasets to be stored
    :param (str) names: attributes of the object that shall be stored (not to be used if no object is given)
    :param names_to_datasets: keyword arguments of explicit datasets with names,
                              if no object is given or additional values shall be stored
    """
    for name in names:
        write_dataset(group, name, obj)
    for name, dataset in names_to_datasets.items():
        write_dataset(group, name, dataset=dataset)


def _write_flat_dataset(group, name, dataset):
    """
    write lists or sets as datasets to the open group of an HDF5 file

    :param (h5py.Group) group: the group in the HDF5 file where the dataset is stored
    :param (str) name: the name under which the dataset is stored
    :param dataset: the dict which will be stored
    """
    # if it is a scalar, it should be stored as an attribute
    if np.isscalar(dataset):
        raise ValueError(ERROR_SCALAR)
    # if there is None stored or the length of the dataset is 0
    if dataset is None:
        np_dataset = np.NaN
    else:
        # if is already transformed into a numpy array assume that it can be stored directly
        if isinstance(dataset, np.ndarray):
            np_dataset = dataset
        else:
            # we do not differ between sets/ tuples/ lists/ other iterables when storing
            if np.iterable(dataset):
                dataset = list(dataset)

            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                # in future numpy will not allow ragged nested arrays anymore
                try:
                    # but e.g. strings of different length are fine
                    np_dataset = np.array(dataset)
                except (np.VisibleDeprecationWarning, ValueError):
                    np_dataset = np.array(dataset, dtype=object)

        if np_dataset.dtype.kind == "U":  # unicode should be converted to ascii for h5py
            np_dataset = np_dataset.astype("S")
        if np_dataset.dtype == object:
            np_dataset = _handle_object_dtype(dataset, np_dataset)

    group.create_dataset(name, data=np_dataset)


def _handle_object_dtype(dataset, np_dataset):
    if not isinstance(dataset, list):
        raise ValueError(ERROR_TYPE.format(type(dataset)))

    if all(np.iterable(x) for x in dataset):
        vlen_array = True
        elements_type = _get_homogeneous_type(itertools.chain.from_iterable(dataset))
    else:
        vlen_array = False
        elements_type = _get_homogeneous_type(dataset)

    if not elements_type:
        raise NotImplementedError(ERROR_NON)
    if elements_type not in np.ScalarType or elements_type == str:
        raise NotImplementedError(ERROR_LIST.format(elements_type))

    if vlen_array:
        vlen_type = h5py.special_dtype(vlen=np.dtype(elements_type))
        np_dataset = np.array([np.array(x, dtype=elements_type) for x in dataset], dtype=vlen_type)
    return np_dataset

def _get_homogeneous_type(sequence):
    type_set = set(type(element) for element in sequence)
    if len(type_set) == 1:
        return type_set.pop()

    if not type_set:
        return None

    # numbers can have different types
    if all(issubclass(t, Real) for t in type_set):
        return float
    return None


def _write_dict_dataset(group, name, dataset):
    """
    write dictionary as dataset to the open group of an HDF5 file
    by splitting it up into 'flat' parts

    :param (h5py.Group) group: the group in the HDF5 file where the dataset is stored
    :param (str) name: the name under which the dataset is stored
    :param dataset: the dict which will be stored
    """
    keys = dataset.keys()
    vals = dataset.values()
    # check if dict homogeneous
    # ignore float keys since not recommendable
    key_type = _get_homogeneous_type(keys)
    if not key_type and len(keys) > 0:
        raise TypeError(ERROR_KEYS)

    if key_type is tuple:
        items_in_tuples = [item for key in keys for item in key]
        item_type = _get_homogeneous_type(items_in_tuples)
        if len(items_in_tuples) > 0 and not item_type:
            raise TypeError(ERROR_ITEMS)

    # create group and write datasets
    dict_group = group.create_group(name)
    if keys:
        _write_flat_dataset(dict_group, KEYS, keys)
        _write_flat_dataset(dict_group, VALS, vals)


def read_dataset(group, name, check_existence=True):
    """
    read dataset from the open group of an HDF5 file

    :param (h5py.Group) group: the group in the HDF5 file where the dataset is stored
    :param (str) name: the name under which the dataset is stored
    :param (bool) check_existence: if True then an error will be thrown if the dataset is not found,
                                   if False None will be returned without error
    :return: the dataset
    """

    if name not in group:
        if check_existence:
            raise ValueError(ERROR_NOT_FOUND.format(name, group.name, group.file.filename))
        return None
    if isinstance(group[name], h5py.Group):
        return _read_dict_dataset(group[name])
    return _read_flat_dataset(group, name)


def _read_flat_dataset(group, name):
    """
    read a flat dataset from the open group of an HDF5 file

    :param (h5py.Group) group: the HDF5 group to read from
    :param (str) name: the name under which the dataset is stored
    :return: the dataset
    """
    dataset = group[name][()]
    if isinstance(dataset, float) and np.isnan(dataset):
        return None
    if dataset.dtype.kind == 'S':
        return dataset.astype('U')
    return dataset


def _read_dict_dataset(group):
    """
    read a dictionary dataset from the open group of an HDF5 file

    :param (h5py.Group) group: the HDF5 group to read from
    """
    # keys and values
    ds_keys = None
    ds_vals = None

    if KEYS in group:
        ds_keys = _read_flat_dataset(group, KEYS)
    if VALS in group:
        ds_vals = _read_flat_dataset(group, VALS)

    if ds_keys is None or ds_vals is None:
        ds_keys = []
        ds_vals = []
    if len(ds_keys) != len(ds_vals):
        raise ValueError(ERROR_MISMATCH.format(group, len(ds_keys), len(ds_vals)))

    if len(ds_keys) == 0:
        return {}

    if isinstance(ds_keys[0], np.ndarray):
        ds_keys = [tuple(key) for key in ds_keys]

    return dict(zip(ds_keys, ds_vals))


def read_datasets(group, *names, check_existence=True):
    """
    read several datasets from the open group of an HDF5 file

    :param (h5py.Group) group: the group in the HDF5 file where the datasets are stored
    :param (str) names: the names under which the datasets are stored
    :param (bool) check_existence: if True then an error will be thrown if a dataset is not found,
                                   if False None will be returned without error
    :return: dictionary with the datasets keyed by names
    """
    return {name: read_dataset(group, name, check_existence) for name in names}
